"""MurmurHash3 hash module."""

from __future__ import absolute_import, division, print_function, \
    unicode_literals

from six.moves import range


def murmur32_py(key, seed=0x0):
    """
    Pure python implementation of murmur32 hash.

    :param key: Key to hash
    :type key: str
    :param seed: Seed to use when hashing
    :type seed: int

    :return: hashed value
    :rtype: int

    """
    key = bytearray(key, 'utf-8')

    def fmix(current_hash):
        """Mix has bytes."""
        current_hash ^= current_hash >> 16
        current_hash = (current_hash * 0x85ebca6b) & 0xFFFFFFFF
        current_hash ^= current_hash >> 13
        current_hash = (current_hash * 0xc2b2ae35) & 0xFFFFFFFF
        current_hash ^= current_hash >> 16
        return current_hash

    length = len(key)
    nblocks = int(length/4)

    hash1 = seed & 0xFFFFFFFF

    calc1 = 0xcc9e2d51
    calc2 = 0x1b873593

    # body
    for block_start in range(0, nblocks * 4, 4):
        # ??? big endian?
        key1 = key[block_start + 3] << 24 | \
             key[block_start + 2] << 16 | \
             key[block_start + 1] << 8 | \
             key[block_start + 0]

        key1 = (calc1 * key1) & 0xFFFFFFFF
        key1 = (key1 << 15 | key1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        key1 = (calc2 * key1) & 0xFFFFFFFF

        hash1 ^= key1
        hash1 = (hash1 << 13 | hash1 >> 19) & 0xFFFFFFFF  # inlined ROTL32
        hash1 = (hash1 * 5 + 0xe6546b64) & 0xFFFFFFFF

    # tail
    tail_index = nblocks * 4
    key1 = 0
    tail_size = length & 3

    if tail_size >= 3:
        key1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        key1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        key1 ^= key[tail_index + 0]

    if tail_size > 0:
        key1 = (key1 * calc1) & 0xFFFFFFFF
        key1 = (key1 << 15 | key1 >> 17) & 0xFFFFFFFF  # inlined ROTL32
        key1 = (key1 * calc2) & 0xFFFFFFFF
        hash1 ^= key1

    unsigned_val = fmix(hash1 ^ length)
    return unsigned_val

def hash128_x64(key, seed):
    """
    Pure python implementation of murmurhash3-128.

    borrowed from: https://github.com/wc-duck/pymmh3/blob/master/pymmh3.py
    """
    key = bytearray(key, 'utf-8')

    def fmix(k):
        k ^= k >> 33
        k  = (k * 0xff51afd7ed558ccd) & 0xFFFFFFFFFFFFFFFF
        k ^= k >> 33
        k  = (k * 0xc4ceb9fe1a85ec53) & 0xFFFFFFFFFFFFFFFF
        k ^= k >> 33
        return k

    length = len(key)
    nblocks = int(length / 16)

    h1 = seed
    h2 = seed

    c1 = 0x87c37b91114253d5
    c2 = 0x4cf5ad432745937f

    #body
    for block_start in range(0, nblocks * 8, 8):
        # ??? big endian?
        k1 = key[2 * block_start + 7] << 56 | \
             key[2 * block_start + 6] << 48 | \
             key[2 * block_start + 5] << 40 | \
             key[2 * block_start + 4] << 32 | \
             key[2 * block_start + 3] << 24 | \
             key[2 * block_start + 2] << 16 | \
             key[2 * block_start + 1] <<  8 | \
             key[2 * block_start + 0]

        k2 = key[2 * block_start + 15] << 56 | \
             key[2 * block_start + 14] << 48 | \
             key[2 * block_start + 13] << 40 | \
             key[2 * block_start + 12] << 32 | \
             key[2 * block_start + 11] << 24 | \
             key[2 * block_start + 10] << 16 | \
             key[2 * block_start + 9] <<  8 | \
             key[2 * block_start + 8]

        k1  = (c1 * k1) & 0xFFFFFFFFFFFFFFFF
        k1  = (k1 << 31 | k1 >> 33) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        k1  = (c2 * k1) & 0xFFFFFFFFFFFFFFFF
        h1 ^= k1

        h1 = (h1 << 27 | h1 >> 37) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        h1 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
        h1 = (h1 * 5 + 0x52dce729) & 0xFFFFFFFFFFFFFFFF

        k2  = (c2 * k2) & 0xFFFFFFFFFFFFFFFF
        k2  = (k2 << 33 | k2 >> 31) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        k2  = (c1 * k2) & 0xFFFFFFFFFFFFFFFF
        h2 ^= k2

        h2 = (h2 << 31 | h2 >> 33) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        h2 = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
        h2 = (h2 * 5 + 0x38495ab5) & 0xFFFFFFFFFFFFFFFF

    #tail
    tail_index = nblocks * 16
    k1 = 0
    k2 = 0
    tail_size = length & 15

    if tail_size >= 15:
        k2 ^= key[tail_index + 14] << 48
    if tail_size >= 14:
        k2 ^= key[tail_index + 13] << 40
    if tail_size >= 13:
        k2 ^= key[tail_index + 12] << 32
    if tail_size >= 12:
        k2 ^= key[tail_index + 11] << 24
    if tail_size >= 11:
        k2 ^= key[tail_index + 10] << 16
    if tail_size >= 10:
        k2 ^= key[tail_index + 9] << 8
    if tail_size >=  9:
        k2 ^= key[tail_index + 8]

    if tail_size > 8:
        k2  = (k2 * c2) & 0xFFFFFFFFFFFFFFFF
        k2  = (k2 << 33 | k2 >> 31) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        k2  = (k2 * c1) & 0xFFFFFFFFFFFFFFFF
        h2 ^= k2

    if tail_size >= 8:
        k1 ^= key[tail_index + 7] << 56
    if tail_size >= 7:
        k1 ^= key[tail_index + 6] << 48
    if tail_size >= 6:
        k1 ^= key[tail_index + 5] << 40
    if tail_size >= 5:
        k1 ^= key[tail_index + 4] << 32
    if tail_size >= 4:
        k1 ^= key[tail_index + 3] << 24
    if tail_size >= 3:
        k1 ^= key[tail_index + 2] << 16
    if tail_size >= 2:
        k1 ^= key[tail_index + 1] << 8
    if tail_size >= 1:
        k1 ^= key[tail_index + 0]

    if tail_size > 0:
        k1  = (k1 * c1) & 0xFFFFFFFFFFFFFFFF
        k1  = (k1 << 31 | k1 >> 33) & 0xFFFFFFFFFFFFFFFF # inlined ROTL64
        k1  = (k1 * c2) & 0xFFFFFFFFFFFFFFFF
        h1 ^= k1

    #finalization
    h1 ^= length
    h2 ^= length

    h1  = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
    h2  = (h1 + h2) & 0xFFFFFFFFFFFFFFFF

    h1  = fmix(h1)
    h2  = fmix(h2)

    h1  = (h1 + h2) & 0xFFFFFFFFFFFFFFFF
    h2  = (h1 + h2) & 0xFFFFFFFFFFFFFFFF

    return [h1, h2]



"""Push notification manager tests."""
#pylint:disable=no-self-use,protected-access
from threading import Thread
from queue import Queue
from splitio.api.auth import APIException
from splitio.push.sse import SSEEvent
from splitio.models.token import Token
from splitio.push.parser import parse_incoming_event, EventType, ControlType, ControlMessage, \
    OccupancyMessage, SplitChangeUpdate, SplitKillUpdate, SegmentChangeUpdate
from splitio.push.processor import MessageProcessor
from splitio.push.status_tracker import PushStatusTracker
from splitio.push.manager import PushManager, _TOKEN_REFRESH_GRACE_PERIOD
from splitio.push.splitsse import SplitSSEClient
from splitio.push.status_tracker import Status
from tests.helpers import Any


class PushManagerTests(object):
    """Parser tests."""

    def test_connection_success(self, mocker):
        """Test the initial status is ok and reset() works as expected."""
        api_mock = mocker.Mock()
        api_mock.authenticate.return_value = Token(True, 'abc', {}, 2000000, 1000000)

        sse_mock = mocker.Mock(spec=SplitSSEClient)
        sse_constructor_mock = mocker.Mock()
        sse_constructor_mock.return_value = sse_mock
        timer_mock = mocker.Mock()
        mocker.patch('splitio.push.manager.Timer', new=timer_mock)
        mocker.patch('splitio.push.manager.SplitSSEClient', new=sse_constructor_mock)
        feedback_loop = Queue()
        manager = PushManager(api_mock, mocker.Mock(), feedback_loop)

        def new_start(*args, **kwargs):  # pylint: disable=unused-argument
            """splitsse.start mock."""
            thread = Thread(target=manager._handle_connection_ready)
            thread.setDaemon(True)
            thread.start()
            return True

        sse_mock.start.side_effect = new_start

        manager.start()
        assert feedback_loop.get() == Status.PUSH_SUBSYSTEM_UP
        assert timer_mock.mock_calls == [
            mocker.call(0, Any()),
            mocker.call().cancel(),
            mocker.call(1000000 - _TOKEN_REFRESH_GRACE_PERIOD, manager._token_refresh),
            mocker.call().setName('TokenRefresh'),
            mocker.call().start()
        ]

    def test_connection_failure(self, mocker):
        """Test the connection fails to be established."""
        api_mock = mocker.Mock()
        api_mock.authenticate.return_value = Token(True, 'abc', {}, 2000000, 1000000)

        sse_mock = mocker.Mock(spec=SplitSSEClient)
        sse_constructor_mock = mocker.Mock()
        sse_constructor_mock.return_value = sse_mock
        timer_mock = mocker.Mock()
        mocker.patch('splitio.push.manager.Timer', new=timer_mock)
        mocker.patch('splitio.push.manager.SplitSSEClient', new=sse_constructor_mock)
        feedback_loop = Queue()
        manager = PushManager(api_mock, mocker.Mock(), feedback_loop)

        def new_start(*args, **kwargs):  # pylint: disable=unused-argument
            """splitsse.start mock."""
            thread = Thread(target=manager._handle_connection_end)
            thread.setDaemon(True)
            thread.start()
            return False

        sse_mock.start.side_effect = new_start

        manager.start()
        assert feedback_loop.get() == Status.PUSH_RETRYABLE_ERROR
        assert timer_mock.mock_calls == [mocker.call(0, Any())]

    def test_push_disabled(self, mocker):
        """Test the initial status is ok and reset() works as expected."""
        api_mock = mocker.Mock()
        api_mock.authenticate.return_value = Token(False, 'abc', {}, 1, 2)

        sse_mock = mocker.Mock(spec=SplitSSEClient)
        sse_constructor_mock = mocker.Mock()
        sse_constructor_mock.return_value = sse_mock
        timer_mock = mocker.Mock()
        mocker.patch('splitio.push.manager.Timer', new=timer_mock)
        mocker.patch('splitio.push.manager.SplitSSEClient', new=sse_constructor_mock)
        feedback_loop = Queue()
        manager = PushManager(api_mock, mocker.Mock(), feedback_loop)
        manager.start()
        assert feedback_loop.get() == Status.PUSH_NONRETRYABLE_ERROR
        assert timer_mock.mock_calls == [mocker.call(0, Any())]
        assert sse_mock.mock_calls == []

    def test_auth_apiexception(self, mocker):
        """Test the initial status is ok and reset() works as expected."""
        api_mock = mocker.Mock()
        api_mock.authenticate.side_effect = APIException('something')

        sse_mock = mocker.Mock(spec=SplitSSEClient)
        sse_constructor_mock = mocker.Mock()
        sse_constructor_mock.return_value = sse_mock
        timer_mock = mocker.Mock()
        mocker.patch('splitio.push.manager.Timer', new=timer_mock)
        mocker.patch('splitio.push.manager.SplitSSEClient', new=sse_constructor_mock)

        feedback_loop = Queue()
        manager = PushManager(api_mock, mocker.Mock(), feedback_loop)
        manager.start()
        assert feedback_loop.get() == Status.PUSH_RETRYABLE_ERROR
        assert timer_mock.mock_calls == [mocker.call(0, Any())]
        assert sse_mock.mock_calls == []

    def test_split_change(self, mocker):
        """Test update-type messages are properly forwarded to the processor."""
        sse_event = SSEEvent('1', EventType.MESSAGE, '', '{}')
        update_message = SplitChangeUpdate('chan', 123, 456)
        parse_event_mock = mocker.Mock(spec=parse_incoming_event)
        parse_event_mock.return_value = update_message
        mocker.patch('splitio.push.manager.parse_incoming_event', new=parse_event_mock)

        processor_mock = mocker.Mock(spec=MessageProcessor)
        mocker.patch('splitio.push.manager.MessageProcessor', new=processor_mock)

        manager = PushManager(mocker.Mock(), mocker.Mock(), mocker.Mock())
        manager._event_handler(sse_event)
        assert parse_event_mock.mock_calls == [mocker.call(sse_event)]
        assert processor_mock.mock_calls == [
            mocker.call(Any()),
            mocker.call().handle(update_message)
        ]

    def test_split_kill(self, mocker):
        """Test update-type messages are properly forwarded to the processor."""
        sse_event = SSEEvent('1', EventType.MESSAGE, '', '{}')
        update_message = SplitKillUpdate('chan', 123, 456, 'some_split', 'off')
        parse_event_mock = mocker.Mock(spec=parse_incoming_event)
        parse_event_mock.return_value = update_message
        mocker.patch('splitio.push.manager.parse_incoming_event', new=parse_event_mock)

        processor_mock = mocker.Mock(spec=MessageProcessor)
        mocker.patch('splitio.push.manager.MessageProcessor', new=processor_mock)

        manager = PushManager(mocker.Mock(), mocker.Mock(), mocker.Mock())
        manager._event_handler(sse_event)
        assert parse_event_mock.mock_calls == [mocker.call(sse_event)]
        assert processor_mock.mock_calls == [
            mocker.call(Any()),
            mocker.call().handle(update_message)
        ]

    def test_segment_change(self, mocker):
        """Test update-type messages are properly forwarded to the processor."""
        sse_event = SSEEvent('1', EventType.MESSAGE, '', '{}')
        update_message = SegmentChangeUpdate('chan', 123, 456, 'some_segment')
        parse_event_mock = mocker.Mock(spec=parse_incoming_event)
        parse_event_mock.return_value = update_message
        mocker.patch('splitio.push.manager.parse_incoming_event', new=parse_event_mock)

        processor_mock = mocker.Mock(spec=MessageProcessor)
        mocker.patch('splitio.push.manager.MessageProcessor', new=processor_mock)

        manager = PushManager(mocker.Mock(), mocker.Mock(), mocker.Mock())
        manager._event_handler(sse_event)
        assert parse_event_mock.mock_calls == [mocker.call(sse_event)]
        assert processor_mock.mock_calls == [
            mocker.call(Any()),
            mocker.call().handle(update_message)
        ]

    def test_control_message(self, mocker):
        """Test control mesage is forwarded to status tracker."""
        sse_event = SSEEvent('1', EventType.MESSAGE, '', '{}')
        control_message = ControlMessage('chan', 123, ControlType.STREAMING_ENABLED)
        parse_event_mock = mocker.Mock(spec=parse_incoming_event)
        parse_event_mock.return_value = control_message
        mocker.patch('splitio.push.manager.parse_incoming_event', new=parse_event_mock)

        status_tracker_mock = mocker.Mock(spec=PushStatusTracker)
        mocker.patch('splitio.push.manager.PushStatusTracker', new=status_tracker_mock)

        manager = PushManager(mocker.Mock(), mocker.Mock(), mocker.Mock())
        manager._event_handler(sse_event)
        assert parse_event_mock.mock_calls == [mocker.call(sse_event)]
        assert status_tracker_mock.mock_calls == [
            mocker.call(),
            mocker.call().handle_control_message(control_message)
        ]

    def test_occupancy_message(self, mocker):
        """Test control mesage is forwarded to status tracker."""
        sse_event = SSEEvent('1', EventType.MESSAGE, '', '{}')
        occupancy_message = OccupancyMessage('[?occupancy=metrics.publishers]control_pri', 123, 2)
        parse_event_mock = mocker.Mock(spec=parse_incoming_event)
        parse_event_mock.return_value = occupancy_message
        mocker.patch('splitio.push.manager.parse_incoming_event', new=parse_event_mock)

        status_tracker_mock = mocker.Mock(spec=PushStatusTracker)
        mocker.patch('splitio.push.manager.PushStatusTracker', new=status_tracker_mock)

        manager = PushManager(mocker.Mock(), mocker.Mock(), mocker.Mock())
        manager._event_handler(sse_event)
        assert parse_event_mock.mock_calls == [mocker.call(sse_event)]
        assert status_tracker_mock.mock_calls == [
            mocker.call(),
            mocker.call().handle_occupancy(occupancy_message)
        ]



"""Redis client wrapper with prefix support."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

from builtins import str
from six import string_types, binary_type
from future.utils import raise_from

try:
    from redis import StrictRedis
    from redis.sentinel import Sentinel
    from redis.exceptions import RedisError
except ImportError:
    def missing_redis_dependencies(*_, **__):
        """Fail if missing dependencies are used."""
        raise NotImplementedError(
            'Missing Redis support dependencies. '
            'Please use `pip install splitio_client[redis]` to install the sdk with redis support'
        )
    StrictRedis = Sentinel = missing_redis_dependencies


def _bytes_to_string(maybe_bytes, encode='utf-8'):
    if type(maybe_bytes).__name__ == 'bytes':
        return str(maybe_bytes, encode)
    return maybe_bytes


class RedisAdapterException(Exception):
    """Exception to be thrown when a redis command fails with an exception."""

    def __init__(self, message, original_exception=None):
        """
        Exception constructor.

        :param message: Custom exception message.
        :type message: str
        :param original_exception: Original exception object.
        :type original_exception: Exception
        """
        Exception.__init__(self, message)
        self._original_exception = original_exception

    @property
    def original_exception(self):
        """Return original exception."""
        return self._original_exception


class SentinelConfigurationException(Exception):
    """Exception to be raised when sentinel config options are incorrect."""

    pass


class RedisAdapter(object):  #pylint: disable=too-many-public-methods
    """
    Instance decorator for Redis clients such as StrictRedis.

    Adds an extra layer handling addition/removal of user prefix when handling
    keys
    """

    def __init__(self, decorated, prefix=None):
        """
        Store the user prefix and the redis client instance.

        :param decorated: Instance of redis cache client to decorate.
        :param prefix: User prefix to add.
        """
        self._prefix = prefix
        self._decorated = decorated

    def _add_prefix(self, k):
        """
        Add a prefix to the contents of k.

        'k' may be:
            - a single key (of type string or unicode in python2, or type string
            in python 3. In which case we simple add a prefix with a dot.
            - a list, in which the prefix is applied to element.
        If no user prefix is stored, the key/list of keys will be returned as is

        :param k: single (string) or list of (list) keys.
        :returns: Key(s) with prefix if applicable
        """
        if self._prefix:
            if isinstance(k, string_types):
                return '{prefix}.{key}'.format(prefix=self._prefix, key=k)
            elif isinstance(k, list) and k:
                if isinstance(k[0], binary_type):
                    return [
                        '{prefix}.{key}'.format(prefix=self._prefix, key=key.decode("utf8"))
                        for key in k
                    ]
                elif isinstance(k[0], string_types):
                    return [
                        '{prefix}.{key}'.format(prefix=self._prefix, key=key)
                        for key in k
                    ]
        else:
            return k

        raise RedisAdapterException(
            "Cannot append prefix correctly. Wrong type for key(s) provided"
        )

    def _remove_prefix(self, k):
        """
        Remove the user prefix from a key before handling it back to the requester.

        Similar to _add_prefix, this class will handle single strings as well
        as lists. If no _prefix is set, the original key/keys will be returned.

        :param k: key(s) whose prefix will be removed.
        :returns: prefix-less key(s)
        """
        if self._prefix:
            if isinstance(k, string_types):
                return k[len(self._prefix)+1:]
            elif isinstance(k, list):
                return [key[len(self._prefix)+1:] for key in k]
        else:
            return k

        raise RedisAdapterException(
            "Cannot remove prefix correctly. Wrong type for key(s) provided"
        )

    # Below starts a list of methods that implement the interface of a standard
    # redis client.

    def keys(self, pattern):
        """Mimic original redis function but using user custom prefix."""
        try:
            return [
                _bytes_to_string(key)
                for key in self._remove_prefix(self._decorated.keys(self._add_prefix(pattern)))
            ]
        except RedisError as exc:
            raise_from(RedisAdapterException('Failed to execute keys operation'), exc)

    def set(self, name, value, *args, **kwargs):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.set(
                self._add_prefix(name), value, *args, **kwargs
            )
        except RedisError as exc:
            raise RedisAdapterException('Failed to execute set operation', exc)

    def get(self, name):
        """Mimic original redis function but using user custom prefix."""
        try:
            return _bytes_to_string(self._decorated.get(self._add_prefix(name)))
        except RedisError as exc:
            raise RedisAdapterException('Error executing get operation', exc)

    def setex(self, name, time, value):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.setex(self._add_prefix(name), time, value)
        except RedisError as exc:
            raise RedisAdapterException('Error executing setex operation', exc)

    def delete(self, *names):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.delete(*self._add_prefix(list(names)))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing delete operation'), exc)

    def exists(self, name):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.exists(self._add_prefix(name))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing exists operation'), exc)

    def lrange(self, key, start, end):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.lrange(self._add_prefix(key), start, end)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing exists operation'), exc)

    def mget(self, names):
        """Mimic original redis function but using user custom prefix."""
        try:
            return [
                _bytes_to_string(item)
                for item in self._decorated.mget(self._add_prefix(names))
            ]
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing mget operation'), exc)

    def smembers(self, name):
        """Mimic original redis function but using user custom prefix."""
        try:
            return [
                _bytes_to_string(item)
                for item in self._decorated.smembers(self._add_prefix(name))
            ]
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing smembers operation'), exc)

    def sadd(self, name, *values):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.sadd(self._add_prefix(name), *values)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing sadd operation'), exc)

    def srem(self, name, *values):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.srem(self._add_prefix(name), *values)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing srem operation'), exc)

    def sismember(self, name, value):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.sismember(self._add_prefix(name), value)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing sismember operation'), exc)

    def eval(self, script, number_of_keys, *keys):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.eval(script, number_of_keys, *self._add_prefix(list(keys)))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing eval operation'), exc)

    def hset(self, name, key, value):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.hset(self._add_prefix(name), key, value)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing hset operation'), exc)

    def hget(self, name, key):
        """Mimic original redis function but using user custom prefix."""
        try:
            return _bytes_to_string(self._decorated.hget(self._add_prefix(name), key))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing hget operation'), exc)

    def incr(self, name, amount=1):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.incr(self._add_prefix(name), amount)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing incr operation'), exc)

    def getset(self, name, value):
        """Mimic original redis function but using user custom prefix."""
        try:
            return _bytes_to_string(self._decorated.getset(self._add_prefix(name), value))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing getset operation'), exc)

    def rpush(self, key, *values):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.rpush(self._add_prefix(key), *values)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing rpush operation'), exc)

    def expire(self, key, value):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.expire(self._add_prefix(key), value)
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing expire operation'), exc)

    def rpop(self, key):
        """Mimic original redis function but using user custom prefix."""
        try:
            return _bytes_to_string(self._decorated.rpop(self._add_prefix(key)))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing rpop operation'), exc)

    def ttl(self, key):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.ttl(self._add_prefix(key))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing ttl operation'), exc)

    def lpop(self, key):
        """Mimic original redis function but using user custom prefix."""
        try:
            return self._decorated.lpop(self._add_prefix(key))
        except RedisError as exc:
            raise_from(RedisAdapterException('Error executing lpop operation'), exc)


def _build_default_client(config):  #pylint: disable=too-many-locals
    """
    Build a redis adapter.

    :param config: Redis configuration properties
    :type config: dict

    :return: A wrapped StrictRedis object
    :rtype: splitio.storage.adapters.redis.RedisAdapter
    """
    host = config.get('redisHost', 'localhost')
    port = config.get('redisPort', 6379)
    database = config.get('redisDb', 0)
    password = config.get('redisPassword', None)
    socket_timeout = config.get('redisSocketTimeout', None)
    socket_connect_timeout = config.get('redisSocketConnectTimeout', None)
    socket_keepalive = config.get('redisSocketKeepalive', None)
    socket_keepalive_options = config.get('redisSocketKeepaliveOptions', None)
    connection_pool = config.get('redisConnectionPool', None)
    unix_socket_path = config.get('redisUnixSocketPath', None)
    encoding = config.get('redisEncoding', 'utf-8')
    encoding_errors = config.get('redisEncodingErrors', 'strict')
    charset = config.get('redisCharset', None)
    errors = config.get('redisErrors', None)
    decode_responses = config.get('redisDecodeResponses', False)
    retry_on_timeout = config.get('redisRetryOnTimeout', False)
    ssl = config.get('redisSsl', False)
    ssl_keyfile = config.get('redisSslKeyfile', None)
    ssl_certfile = config.get('redisSslCertfile', None)
    ssl_cert_reqs = config.get('redisSslCertReqs', None)
    ssl_ca_certs = config.get('redisSslCaCerts', None)
    max_connections = config.get('redisMaxConnections', None)
    prefix = config.get('redisPrefix')

    redis = StrictRedis(
        host=host,
        port=port,
        db=database,
        password=password,
        socket_timeout=socket_timeout,
        socket_connect_timeout=socket_connect_timeout,
        socket_keepalive=socket_keepalive,
        socket_keepalive_options=socket_keepalive_options,
        connection_pool=connection_pool,
        unix_socket_path=unix_socket_path,
        encoding=encoding,
        encoding_errors=encoding_errors,
        charset=charset,
        errors=errors,
        decode_responses=decode_responses,
        retry_on_timeout=retry_on_timeout,
        ssl=ssl,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        ssl_cert_reqs=ssl_cert_reqs,
        ssl_ca_certs=ssl_ca_certs,
        max_connections=max_connections
    )
    return RedisAdapter(redis, prefix=prefix)


def _build_sentinel_client(config):  #pylint: disable=too-many-locals
    """
    Build a redis client with sentinel replication.

    :param config: Redis configuration properties.
    :type config: dict

    :return: A Wrapped redis-sentinel client
    :rtype: splitio.storage.adapters.redis.RedisAdapter
    """
    sentinels = config.get('redisSentinels')

    if config.get('redisSsl', False):
        raise SentinelConfigurationException('Redis Sentinel cannot be used with SSL/TLS.')

    if sentinels is None:
        raise SentinelConfigurationException('redisSentinels must be specified.')
    if not isinstance(sentinels, list):
        raise SentinelConfigurationException('Sentinels must be an array of elements in the form of'
                                             ' [(ip, port)].')
    if not sentinels:
        raise SentinelConfigurationException('It must be at least one sentinel.')
    if not all(isinstance(s, tuple) for s in sentinels):
        raise SentinelConfigurationException('Sentinels must respect the tuple structure'
                                             '[(ip, port)].')

    master_service = config.get('redisMasterService')

    if master_service is None:
        raise SentinelConfigurationException('redisMasterService must be specified.')

    database = config.get('redisDb', 0)
    password = config.get('redisPassword', None)
    socket_timeout = config.get('redisSocketTimeout', None)
    socket_connect_timeout = config.get('redisSocketConnectTimeout', None)
    socket_keepalive = config.get('redisSocketKeepalive', None)
    socket_keepalive_options = config.get('redisSocketKeepaliveOptions', None)
    connection_pool = config.get('redisConnectionPool', None)
    encoding = config.get('redisEncoding', 'utf-8')
    encoding_errors = config.get('redisEncodingErrors', 'strict')
    decode_responses = config.get('redisDecodeResponses', False)
    retry_on_timeout = config.get('redisRetryOnTimeout', False)
    max_connections = config.get('redisMaxConnections', None)
    prefix = config.get('redisPrefix')

    sentinel = Sentinel(
        sentinels,
        db=database,
        password=password,
        socket_timeout=socket_timeout,
        socket_connect_timeout=socket_connect_timeout,
        socket_keepalive=socket_keepalive,
        socket_keepalive_options=socket_keepalive_options,
        connection_pool=connection_pool,
        encoding=encoding,
        encoding_errors=encoding_errors,
        decode_responses=decode_responses,
        retry_on_timeout=retry_on_timeout,
        max_connections=max_connections
    )

    redis = sentinel.master_for(master_service)
    return RedisAdapter(redis, prefix=prefix)


def build(config):
    """
    Build a redis storage according to the configuration received.

    :param config: SDK Configuration parameters with redis properties.
    :type config: dict.

    :return: A redis client.
    :rtype: splitio.storage.adapters.redis.RedisAdapter
    """
    if 'redisSentinels' in config:
        return _build_sentinel_client(config)
    return _build_default_client(config)



from __future__ import absolute_import, division, print_function, \
    unicode_literals

from splitio.client.factory import get_factory
from splitio.client.key import Key
from splitio.version import __version__



"""Split evaluator module."""
import logging
import six
from splitio.models.grammar.condition import ConditionType
from splitio.models.impressions import Label


CONTROL = 'control'


_LOGGER = logging.getLogger(__name__)


class Evaluator(object):  # pylint: disable=too-few-public-methods
    """Split Evaluator class."""

    def __init__(self, split_storage, segment_storage, splitter):
        """
        Construct a Evaluator instance.

        :param split_storage: Split storage.
        :type split_storage: splitio.storage.SplitStorage

        :param split_storage: Storage storage.
        :type split_storage: splitio.storage.SegmentStorage
        """
        self._split_storage = split_storage
        self._segment_storage = segment_storage
        self._splitter = splitter

    def _evaluate_treatment(self, feature, matching_key, bucketing_key, attributes, split):
        """
        Evaluate the user submitted data against a feature and return the resulting treatment.

        :param feature: The feature for which to get the treatment
        :type feature:  str

        :param matching_key: The matching_key for which to get the treatment
        :type matching_key: str

        :param bucketing_key: The bucketing_key for which to get the treatment
        :type bucketing_key: str

        :param attributes: An optional dictionary of attributes
        :type attributes: dict

        :param split: Split object
        :type attributes: splitio.models.splits.Split|None

        :return: The treatment for the key and split
        :rtype: object
        """
        label = ''
        _treatment = CONTROL
        _change_number = -1

        if split is None:
            _LOGGER.warning('Unknown or invalid feature: %s', feature)
            label = Label.SPLIT_NOT_FOUND
        else:
            _change_number = split.change_number
            if split.killed:
                label = Label.KILLED
                _treatment = split.default_treatment
            else:
                treatment, label = self._get_treatment_for_split(
                    split,
                    matching_key,
                    bucketing_key,
                    attributes
                )
                if treatment is None:
                    label = Label.NO_CONDITION_MATCHED
                    _treatment = split.default_treatment
                else:
                    _treatment = treatment

        return {
            'treatment': _treatment,
            'configurations': split.get_configurations_for(_treatment) if split else None,
            'impression': {
                'label': label,
                'change_number': _change_number
            }
        }

    def evaluate_feature(self, feature, matching_key, bucketing_key, attributes=None):
        """
        Evaluate the user submitted data against a feature and return the resulting treatment.

        :param feature: The feature for which to get the treatment
        :type feature:  str

        :param matching_key: The matching_key for which to get the treatment
        :type matching_key: str

        :param bucketing_key: The bucketing_key for which to get the treatment
        :type bucketing_key: str

        :param attributes: An optional dictionary of attributes
        :type attributes: dict

        :return: The treatment for the key and split
        :rtype: object
        """
        # Fetching Split definition
        split = self._split_storage.get(feature)

        # Calling evaluation
        evaluation = self._evaluate_treatment(feature, matching_key,
                                              bucketing_key, attributes, split)

        return evaluation

    def evaluate_features(self, features, matching_key, bucketing_key, attributes=None):
        """
        Evaluate the user submitted data against multiple features and return the resulting
        treatment.

        :param features: The features for which to get the treatments
        :type feature:  list(str)

        :param matching_key: The matching_key for which to get the treatment
        :type matching_key: str

        :param bucketing_key: The bucketing_key for which to get the treatment
        :type bucketing_key: str

        :param attributes: An optional dictionary of attributes
        :type attributes: dict

        :return: The treatments for the key and splits
        :rtype: object
        """
        return {
            feature: self._evaluate_treatment(feature, matching_key,
                                              bucketing_key, attributes, split)
            for (feature, split) in six.iteritems(self._split_storage.fetch_many(features))
        }

    def _get_treatment_for_split(self, split, matching_key, bucketing_key, attributes=None):
        """
        Evaluate the feature considering the conditions.

        If there is a match, it will return the condition and the label.
        Otherwise, it will return (None, None)

        :param split: The split for which to get the treatment
        :type split: Split

        :param matching_key: The key for which to get the treatment
        :type key: str

        :param bucketing_key: The key for which to get the treatment
        :type key: str

        :param attributes: An optional dictionary of attributes
        :type attributes: dict

        :return: The resulting treatment and label
        :rtype: tuple
        """
        if bucketing_key is None:
            bucketing_key = matching_key

        roll_out = False

        context = {
            'segment_storage': self._segment_storage,
            'evaluator': self,
            'bucketing_key': bucketing_key
        }

        for condition in split.conditions:
            if (not roll_out and
                    condition.condition_type == ConditionType.ROLLOUT):
                if split.traffic_allocation < 100:
                    bucket = self._splitter.get_bucket(
                        bucketing_key,
                        split.traffic_allocation_seed,
                        split.algo
                    )
                    if bucket > split.traffic_allocation:
                        return split.default_treatment, Label.NOT_IN_SPLIT
                roll_out = True

            condition_matches = condition.matches(
                matching_key,
                attributes=attributes,
                context=context
            )

            if condition_matches:
                return self._splitter.get_treatment(
                    bucketing_key,
                    split.seed,
                    condition.partitions,
                    split.algo
                ), condition.label

        # No condition matches
        return None, None



"""Split Worker tests."""

import threading
import time
import pytest

from splitio.api.client import HttpResponse
from splitio.api import APIException
from splitio.storage import EventStorage
from splitio.models.events import Event
from splitio.sync.event import EventSynchronizer


class EventsSynchronizerTests(object):
    """Events synchronizer test cases."""

    def test_synchronize_events_error(self, mocker):
        storage = mocker.Mock(spec=EventStorage)
        storage.pop_many.return_value = [
            Event('key1', 'user', 'purchase', 5.3, 123456, None),
            Event('key2', 'user', 'purchase', 5.3, 123456, None),
        ]

        api = mocker.Mock()

        def run(x):
            raise APIException("something broke")

        api.flush_events.side_effect = run
        event_synchronizer = EventSynchronizer(api, storage, 5)
        event_synchronizer.synchronize_events()
        assert event_synchronizer._failed.qsize() == 2

    def test_synchronize_events_empty(self, mocker):
        storage = mocker.Mock(spec=EventStorage)
        storage.pop_many.return_value = []

        api = mocker.Mock()

        def run(x):
            run._called += 1

        run._called = 0
        api.flush_events.side_effect = run
        event_synchronizer = EventSynchronizer(api, storage, 5)
        event_synchronizer.synchronize_events()
        assert run._called == 0

    def test_synchronize_impressions(self, mocker):
        storage = mocker.Mock(spec=EventStorage)
        storage.pop_many.return_value = [
            Event('key1', 'user', 'purchase', 5.3, 123456, None),
            Event('key2', 'user', 'purchase', 5.3, 123456, None),
        ]

        api = mocker.Mock()

        def run(x):
            run._called += 1
            return HttpResponse(200, '')

        api.flush_events.side_effect = run
        run._called = 0

        event_synchronizer = EventSynchronizer(api, storage, 5)
        event_synchronizer.synchronize_events()
        assert run._called == 1
        assert event_synchronizer._failed.qsize() == 0



"""Misc decorators."""
import sys

from abc import abstractmethod, abstractproperty

def abstract_property(func):
    """
    Python2/3 compatible abstract property decorator.

    :param func: method to decorate
    :type func: callable

    :returns: decorated function
    :rtype: callable
    """
    return (property(abstractmethod(func)) if sys.version_info > (3, 3)
            else abstractproperty(func))



"""Synchronizer tests."""

import pytest

from splitio.sync.synchronizer import Synchronizer, SplitTasks, SplitSynchronizers
from splitio.tasks.split_sync import SplitSynchronizationTask
from splitio.tasks.segment_sync import SegmentSynchronizationTask
from splitio.tasks.impressions_sync import ImpressionsSyncTask, ImpressionsCountSyncTask
from splitio.tasks.events_sync import EventsSyncTask
from splitio.tasks.telemetry_sync import TelemetrySynchronizationTask
from splitio.sync.split import SplitSynchronizer
from splitio.sync.segment import SegmentSynchronizer
from splitio.sync.impression import ImpressionSynchronizer, ImpressionsCountSynchronizer
from splitio.sync.event import EventSynchronizer
from splitio.sync.telemetry import TelemetrySynchronizer
from splitio.storage import SegmentStorage, SplitStorage
from splitio.api import APIException
from splitio.models.splits import Split
from splitio.models.segments import Segment


class SynchronizerTests(object):
    def test_sync_all_failed_splits(self, mocker):
        api = mocker.Mock()
        storage = mocker.Mock()

        def run(x):
            raise APIException("something broke")
        api.fetch_splits.side_effect = run

        split_sync = SplitSynchronizer(api, storage)
        split_synchronizers = SplitSynchronizers(split_sync, mocker.Mock(), mocker.Mock(),
                                                 mocker.Mock(), mocker.Mock(), mocker.Mock())
        sychronizer = Synchronizer(split_synchronizers, mocker.Mock(spec=SplitTasks))

        with pytest.raises(APIException):
            sychronizer.synchronize_splits(None)
        with pytest.raises(APIException):
            sychronizer.sync_all()

    def test_sync_all_failed_segments(self, mocker):
        api = mocker.Mock()
        storage = mocker.Mock()
        split_storage = mocker.Mock(spec=SplitStorage)
        split_storage.get_segment_names.return_value = ['segmentA']
        split_sync = mocker.Mock(spec=SplitSynchronizer)
        split_sync.synchronize_splits.return_value = None

        def run(x, y):
            raise APIException("something broke")
        api.fetch_segment.side_effect = run

        segment_sync = SegmentSynchronizer(api, split_storage, storage)
        split_synchronizers = SplitSynchronizers(split_sync, segment_sync, mocker.Mock(),
                                                 mocker.Mock(), mocker.Mock(), mocker.Mock())
        sychronizer = Synchronizer(split_synchronizers, mocker.Mock(spec=SplitTasks))

        with pytest.raises(RuntimeError):
            sychronizer.sync_all()
        assert not sychronizer._synchronize_segments()

    splits = [{
        'changeNumber': 123,
        'trafficTypeName': 'user',
        'name': 'some_name',
        'trafficAllocation': 100,
        'trafficAllocationSeed': 123456,
        'seed': 321654,
        'status': 'ACTIVE',
        'killed': False,
        'defaultTreatment': 'off',
        'algo': 2,
        'conditions': []
    }]

    def test_sync_all(self, mocker):
        split_storage = mocker.Mock(spec=SplitStorage)
        split_storage.get_change_number.return_value = 123
        split_storage.get_segment_names.return_value = ['segmentA']
        split_api = mocker.Mock()
        split_api.fetch_splits.return_value = {'splits': self.splits, 'since': 123,
                                               'till': 123}
        split_sync = SplitSynchronizer(split_api, split_storage)

        segment_storage = mocker.Mock(spec=SegmentStorage)
        segment_storage.get_change_number.return_value = 123
        segment_api = mocker.Mock()
        segment_api.fetch_segment.return_value = {'name': 'segmentA', 'added': ['key1', 'key2',
                                                  'key3'], 'removed': [], 'since': 123, 'till': 123}
        segment_sync = SegmentSynchronizer(segment_api, split_storage, segment_storage)

        split_synchronizers = SplitSynchronizers(split_sync, segment_sync, mocker.Mock(),
                                                 mocker.Mock(), mocker.Mock(), mocker.Mock())

        synchronizer = Synchronizer(split_synchronizers, mocker.Mock(spec=SplitTasks))
        synchronizer.sync_all()

        inserted_split = split_storage.put.mock_calls[0][1][0]
        assert isinstance(inserted_split, Split)
        assert inserted_split.name == 'some_name'

        inserted_segment = segment_storage.update.mock_calls[0][1]
        assert inserted_segment[0] == 'segmentA'
        assert inserted_segment[1] == ['key1', 'key2', 'key3']
        assert inserted_segment[2] == []

    def test_start_periodic_fetching(self, mocker):
        split_task = mocker.Mock(spec=SplitSynchronizationTask)
        segment_task = mocker.Mock(spec=SegmentSynchronizationTask)
        split_tasks = SplitTasks(split_task, segment_task, mocker.Mock(), mocker.Mock(),
                                 mocker.Mock(), mocker.Mock())
        synchronizer = Synchronizer(mocker.Mock(spec=SplitSynchronizers), split_tasks)
        synchronizer.start_periodic_fetching()

        assert len(split_task.start.mock_calls) == 1
        assert len(segment_task.start.mock_calls) == 1

    def test_stop_periodic_fetching(self, mocker):
        split_task = mocker.Mock(spec=SplitSynchronizationTask)
        segment_task = mocker.Mock(spec=SegmentSynchronizationTask)
        segment_sync = mocker.Mock(spec=SegmentSynchronizer)
        split_synchronizers = SplitSynchronizers(mocker.Mock(), segment_sync, mocker.Mock(),
                                                 mocker.Mock(), mocker.Mock(), mocker.Mock())
        split_tasks = SplitTasks(split_task, segment_task, mocker.Mock(), mocker.Mock(),
                                 mocker.Mock(), mocker.Mock())
        synchronizer = Synchronizer(split_synchronizers, split_tasks)
        synchronizer.stop_periodic_fetching()

        assert len(split_task.stop.mock_calls) == 1
        assert len(segment_task.stop.mock_calls) == 1
        assert len(segment_sync.shutdown.mock_calls) == 0

    def test_start_periodic_data_recording(self, mocker):
        impression_task = mocker.Mock(spec=ImpressionsSyncTask)
        impression_count_task = mocker.Mock(spec=ImpressionsCountSyncTask)
        event_task = mocker.Mock(spec=EventsSyncTask)
        telemetry_task = mocker.Mock(spec=TelemetrySynchronizationTask)
        split_tasks = SplitTasks(mocker.Mock(), mocker.Mock(), impression_task, event_task,
                                 telemetry_task, impression_count_task)
        synchronizer = Synchronizer(mocker.Mock(spec=SplitSynchronizers), split_tasks)
        synchronizer.start_periodic_data_recording()

        assert len(impression_task.start.mock_calls) == 1
        assert len(impression_count_task.start.mock_calls) == 1
        assert len(event_task.start.mock_calls) == 1
        assert len(telemetry_task.start.mock_calls) == 1

    def test_stop_periodic_data_recording(self, mocker):

        def stop_mock(event):
            event.set()
            return

        def stop_mock_2():
            return

        impression_task = mocker.Mock(spec=ImpressionsSyncTask)
        impression_task.stop.side_effect = stop_mock
        impression_count_task = mocker.Mock(spec=ImpressionsCountSyncTask)
        impression_count_task.stop.side_effect = stop_mock
        event_task = mocker.Mock(spec=EventsSyncTask)
        event_task.stop.side_effect = stop_mock
        telemetry_task = mocker.Mock(spec=TelemetrySynchronizationTask)
        telemetry_task.stop.side_effect = stop_mock_2
        split_tasks = SplitTasks(mocker.Mock(), mocker.Mock(), impression_task, event_task,
                                 telemetry_task, impression_count_task)
        synchronizer = Synchronizer(mocker.Mock(spec=SplitSynchronizers), split_tasks)
        synchronizer.stop_periodic_data_recording(True)

        assert len(impression_task.stop.mock_calls) == 1
        assert len(impression_count_task.stop.mock_calls) == 1
        assert len(event_task.stop.mock_calls) == 1
        assert len(telemetry_task.stop.mock_calls) == 1

    def test_shutdown(self, mocker):

        def stop_mock(event):
            event.set()
            return

        def stop_mock_2():
            return

        split_task = mocker.Mock(spec=SplitSynchronizationTask)
        split_task.stop.side_effect = stop_mock_2
        segment_task = mocker.Mock(spec=SegmentSynchronizationTask)
        segment_task.stop.side_effect = stop_mock_2
        impression_task = mocker.Mock(spec=ImpressionsSyncTask)
        impression_task.stop.side_effect = stop_mock
        impression_count_task = mocker.Mock(spec=ImpressionsCountSyncTask)
        impression_count_task.stop.side_effect = stop_mock
        event_task = mocker.Mock(spec=EventsSyncTask)
        event_task.stop.side_effect = stop_mock
        telemetry_task = mocker.Mock(spec=TelemetrySynchronizationTask)
        telemetry_task.stop.side_effect = stop_mock_2

        segment_sync = mocker.Mock(spec=SegmentSynchronizer)

        split_synchronizers = SplitSynchronizers(mocker.Mock(), segment_sync, mocker.Mock(),
                                                 mocker.Mock(), mocker.Mock(), mocker.Mock())
        split_tasks = SplitTasks(split_task, segment_task, impression_task, event_task,
                                 telemetry_task, impression_count_task)
        synchronizer = Synchronizer(split_synchronizers, split_tasks)
        synchronizer.shutdown(True)

        assert len(split_task.stop.mock_calls) == 1
        assert len(segment_task.stop.mock_calls) == 1
        assert len(segment_sync.shutdown.mock_calls) == 1
        assert len(impression_task.stop.mock_calls) == 1
        assert len(impression_count_task.stop.mock_calls) == 1
        assert len(event_task.stop.mock_calls) == 1
        assert len(telemetry_task.stop.mock_calls) == 1



"""Events syncrhonization task."""
from __future__ import absolute_import, division, print_function, \
    unicode_literals

import logging

from six.moves import queue
from splitio.api import APIException
from splitio.tasks import BaseSynchronizationTask
from splitio.tasks.util.asynctask import AsyncTask


_LOGGER = logging.getLogger(__name__)


class EventsSyncTask(BaseSynchronizationTask):
    """Events synchronization task uses an asynctask.AsyncTask to send events."""

    def __init__(self, synchronize_events, period):
        """
        Class constructor.

        :param synchronize_events: Events Api object to send data to the backend
        :type synchronize_events: splitio.api.events.EventsAPI
        :param period: How many seconds to wait between subsequent event pushes to the BE.
        :type period: int

        """
        self._period = period
        self._task = AsyncTask(synchronize_events, self._period, on_stop=synchronize_events)

    def start(self):
        """Start executing the events synchronization task."""
        self._task.start()

    def stop(self, event=None):
        """Stop executing the events synchronization task."""
        self._task.stop(event)

    def flush(self):
        """Flush events in storage."""
        _LOGGER.debug('Forcing flush execution for events')
        self._task.force_execution()

    def is_running(self):
        """
        Return whether the task is running or not.

        :return: True if the task is running. False otherwise.
        :rtype: bool
        """
        return self._task.running()



"""Split Worker tests."""

import threading
import time
import pytest

from splitio.api.client import HttpResponse
from splitio.api import APIException
from splitio.storage import TelemetryStorage
from splitio.sync.telemetry import TelemetrySynchronizer
from splitio.api.telemetry import TelemetryAPI


class TelemetrySynchronizerTests(object):
    """Telemetry synchronizer test cases."""

    def test_synchronize_impressions(self, mocker):
        """Test normal behaviour of sync task."""
        api = mocker.Mock(spec=TelemetryAPI)
        storage = mocker.Mock(spec=TelemetryStorage)
        storage.pop_latencies.return_value = {
            'some_latency1': [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'some_latency2': [0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        }
        storage.pop_gauges.return_value = {
            'gauge1': 123,
            'gauge2': 456
        }
        storage.pop_counters.return_value = {
            'counter1': 1,
            'counter2': 5
        }
        telemetry_synchronizer = TelemetrySynchronizer(api, storage)
        telemetry_synchronizer.synchronize_telemetry()

        assert mocker.call() in storage.pop_latencies.mock_calls
        assert mocker.call() in storage.pop_counters.mock_calls
        assert mocker.call() in storage.pop_gauges.mock_calls

        assert mocker.call({
            'some_latency1': [1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            'some_latency2': [0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        }) in api.flush_latencies.mock_calls

        assert mocker.call({
            'gauge1': 123,
            'gauge2': 456
        }) in api.flush_gauges.mock_calls

        assert mocker.call({
            'counter1': 1,
            'counter2': 5
        }) in api.flush_counters.mock_calls



"""An SSE client wrapper to be used with split endpoint."""
import logging
import threading
from enum import Enum
import six
from splitio.push.sse import SSEClient, SSE_EVENT_ERROR
from splitio.util.threadutil import EventGroup


_LOGGER = logging.getLogger(__name__)


class SplitSSEClient(object):  # pylint: disable=too-many-instance-attributes
    """Split streaming endpoint SSE client."""

    KEEPALIVE_TIMEOUT = 70

    class _Status(Enum):
        IDLE = 0
        CONNECTING = 1
        ERRORED = 2
        CONNECTED = 3

    def __init__(self, event_callback, first_event_callback=None,
                 connection_closed_callback=None, base_url='https://streaming.split.io'):
        """
        Construct a split sse client.

        :param callback: fuction to call when an event is received.
        :type callback: callable

        :param first_event_callback: function to call when the first event is received.
        :type first_event_callback: callable

        :param connection_closed_callback: funciton to call when the connection ends.
        :type connection_closed_callback: callable

        :param base_url: scheme + :// + host
        :type base_url: str
        """
        self._client = SSEClient(self._raw_event_handler)
        self._callback = event_callback
        self._on_connected = first_event_callback
        self._on_disconnected = connection_closed_callback
        self._base_url = base_url
        self._status = SplitSSEClient._Status.IDLE
        self._sse_first_event = None
        self._sse_connection_closed = None

    def _raw_event_handler(self, event):
        """
        Handle incoming raw sse event.

        :param event: Incoming raw sse event.
        :type event: splitio.push.sse.SSEEvent
        """
        if self._status == SplitSSEClient._Status.CONNECTING:
            self._status = SplitSSEClient._Status.CONNECTED if event.event != SSE_EVENT_ERROR \
                else SplitSSEClient._Status.ERRORED
            self._sse_first_event.set()
            if self._on_connected is not None:
                self._on_connected()

        if event.data is not None:
            self._callback(event)

    @staticmethod
    def _format_channels(channels):
        """
        Format channels into a list from the raw object retrieved in the token.

        :param channels: object as extracted from the JWT capabilities.
        :type channels: dict[str,list[str]]

        :returns: channels as a list of strings.
        :rtype: list[str]
        """
        regular = [k for (k, v) in six.iteritems(channels) if v == ['subscribe']]
        occupancy = ['[?occupancy=metrics.publishers]' + k
                     for (k, v) in six.iteritems(channels)
                     if 'channel-metadata:publishers' in v]
        return regular + occupancy

    def _build_url(self, token):
        """
        Build the url to connect to and return it as a string.

        :param token: (parsed) JWT
        :type token: splitio.models.token.Token

        :returns: true if the connection was successful. False otherwise.
        :rtype: bool
        """
        return '{base}/event-stream?v=1.1&accessToken={token}&channels={channels}'.format(
            base=self._base_url,
            token=token.token,
            channels=','.join(self._format_channels(token.channels)))

    def start(self, token):
        """
        Open a connection to start listening for events.

        :param token: (parsed) JWT
        :type token: splitio.models.token.Token

        :returns: true if the connection was successful. False otherwise.
        :rtype: bool
        """
        if self._status != SplitSSEClient._Status.IDLE:
            raise Exception('SseClient already started.')

        self._status = SplitSSEClient._Status.CONNECTING

        event_group = EventGroup()
        self._sse_first_event = event_group.make_event()
        self._sse_connection_closed = event_group.make_event()

        def connect(url):
            """Connect to sse in a blocking manner."""
            try:
                self._client.start(url, timeout=self.KEEPALIVE_TIMEOUT)
            finally:
                self._status = SplitSSEClient._Status.IDLE
                self._sse_connection_closed.set()
                self._on_disconnected()

        url = self._build_url(token)
        task = threading.Thread(target=connect, name='SSEConnection', args=(url,))
        task.setDaemon(True)
        task.start()
        event_group.wait()
        return self._status == SplitSSEClient._Status.CONNECTED

    def stop(self, blocking=False, timeout=None):
        """Abort the ongoing connection."""
        if self._status == SplitSSEClient._Status.IDLE:
            _LOGGER.warn('sse already closed. ignoring')
            return

        self._client.shutdown()
        if blocking:
            self._sse_connection_closed.wait(timeout)



"""Splits module."""
from __future__ import absolute_import, division, print_function, unicode_literals

from enum import Enum
from collections import namedtuple
from future.utils import python_2_unicode_compatible

from splitio.models.grammar import condition


SplitView = namedtuple(
    'SplitView',
    ['name', 'traffic_type', 'killed', 'treatments', 'change_number', 'configs']
)


class Status(Enum):
    """Split status."""

    ACTIVE = "ACTIVE"
    ARCHIVED = "ARCHIVED"


class HashAlgorithm(Enum):
    """Hash algorithm names."""

    LEGACY = 1
    MURMUR = 2


class Split(object):  # pylint: disable=too-many-instance-attributes
    """Split model object."""

    def __init__(  # pylint: disable=too-many-arguments
            self,
            name,
            seed,
            killed,
            default_treatment,
            traffic_type_name,
            status,
            change_number,
            conditions=None,
            algo=None,
            traffic_allocation=None,
            traffic_allocation_seed=None,
            configurations=None
    ):
        """
        Class constructor.

        :param name: Name of the feature
        :type name: unicode
        :param seed: Seed
        :type seed: int
        :param killed: Whether the split is killed or not
        :type killed: bool
        :param default_treatment: Default treatment for the split
        :type default_treatment: str
        :param conditions: Set of conditions to test
        :type conditions: list
        :param algo: Hash algorithm to use when splitting.
        :type algo: HashAlgorithm
        :param traffic_allocation: Percentage of traffic to consider.
        :type traffic_allocation: int
        :pram traffic_allocation_seed: Seed used to hash traffic allocation.
        :type traffic_allocation_seed: int
        """
        self._name = name
        self._seed = seed
        self._killed = killed
        self._default_treatment = default_treatment
        self._traffic_type_name = traffic_type_name
        try:
            self._status = Status(status)
        except ValueError:
            self._status = Status.ARCHIVED

        self._change_number = change_number
        self._conditions = conditions if conditions is not None else []

        if traffic_allocation is None:
            self._traffic_allocation = 100
        elif traffic_allocation >= 0 and traffic_allocation <= 100:
            self._traffic_allocation = traffic_allocation
        else:
            self._traffic_allocation = 100

        self._traffic_allocation_seed = traffic_allocation_seed
        try:
            self._algo = HashAlgorithm(algo)
        except ValueError:
            self._algo = HashAlgorithm.LEGACY

        self._configurations = configurations

    @property
    def name(self):
        """Return name."""
        return self._name

    @property
    def seed(self):
        """Return seed."""
        return self._seed

    @property
    def algo(self):
        """Return hash algorithm."""
        return self._algo

    @property
    def killed(self):
        """Return whether the split has been killed."""
        return self._killed

    @property
    def default_treatment(self):
        """Return the default treatment."""
        return self._default_treatment

    @property
    def traffic_type_name(self):
        """Return the traffic type of the split."""
        return self._traffic_type_name

    @property
    def status(self):
        """Return the status of the split."""
        return self._status

    @property
    def change_number(self):
        """Return the change number of the split."""
        return self._change_number

    @property
    def conditions(self):
        """Return the condition list of the split."""
        return self._conditions

    @property
    def traffic_allocation(self):
        """Return the traffic allocation percentage of the split."""
        return self._traffic_allocation

    @property
    def traffic_allocation_seed(self):
        """Return the traffic allocation seed of the split."""
        return self._traffic_allocation_seed

    def get_configurations_for(self, treatment):
        """Return the mapping of treatments to configurations."""
        return self._configurations.get(treatment) if self._configurations else None

    def get_segment_names(self):
        """
        Return a list of segment names referenced in all matchers from this split.

        :return: List of segment names.
        :rtype: list(string)
        """
        return [name for cond in self.conditions for name in cond.get_segment_names()]

    def to_json(self):
        """Return a JSON representation of this split."""
        return {
            'changeNumber': self.change_number,
            'trafficTypeName': self.traffic_type_name,
            'name': self.name,
            'trafficAllocation': self.traffic_allocation,
            'trafficAllocationSeed': self.traffic_allocation_seed,
            'seed': self.seed,
            'status': self.status.value,
            'killed': self.killed,
            'defaultTreatment': self.default_treatment,
            'algo': self.algo.value,
            'conditions': [c.to_json() for c in self.conditions],
            'configurations': self._configurations
        }

    def to_split_view(self):
        """
        Return a SplitView for the manager.

        :return: A portion of the split useful for inspecting by the user.
        :rtype: SplitView
        """
        return SplitView(
            self.name,
            self.traffic_type_name,
            self.killed,
            list(set(part.treatment for cond in self.conditions for part in cond.partitions)),
            self.change_number,
            self._configurations if self._configurations is not None else {}
        )

    def local_kill(self, default_treatment, change_number):
        """
        Perform split kill.

        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        self._default_treatment = default_treatment
        self._change_number = change_number
        self._killed = True

    @python_2_unicode_compatible
    def __str__(self):
        """Return string representation."""
        return 'name: {name}, seed: {seed}, killed: {killed}, ' \
               'default treatment: {default_treatment}, ' \
               'conditions: {conditions}'.format(
                   name=self._name, seed=self._seed, killed=self._killed,
                   default_treatment=self._default_treatment,
                   conditions=','.join(map(str, self._conditions))
               )


def from_raw(raw_split):
    """
    Parse a split from a JSON portion of splitChanges.

    :param raw_split: JSON object extracted from a splitChange's split array (splitChanges response)
    :type raw_split: dict

    :return: A parsed Split object capable of performing evaluations.
    :rtype: Split
    """
    return Split(
        raw_split['name'],
        raw_split['seed'],
        raw_split['killed'],
        raw_split['defaultTreatment'],
        raw_split['trafficTypeName'],
        raw_split['status'],
        raw_split['changeNumber'],
        [condition.from_raw(c) for c in raw_split['conditions']],
        raw_split.get('algo'),
        traffic_allocation=raw_split.get('trafficAllocation'),
        traffic_allocation_seed=raw_split.get('trafficAllocationSeed'),
        configurations=raw_split.get('configurations')
    )



"""Simple test-and-set LRU Cache."""
import threading


DEFAULT_MAX_SIZE = 5000


class SimpleLruCache(object):  #pylint: disable=too-many-instance-attributes
    """
    Key/Value local memory cache. with expiration & LRU eviction.

    LRU double-linked-list format:

    {
        'key1'---------------------------------------------------------------
        'key2'------------------------------------                           |
        'key3'------------                        |                          |
    }                     |                       |                          |
                          V                       V                          V
                     || MRU  ||  -previous-> ||   X  || ... -previous-> || LRU  || -previous-> None
    None <---next--- || node ||  <---next--- || node || ... <---next--- || node ||
    """

    class _Node(object):  #pylint: disable=too-few-public-methods
        """Links to previous an next items in the circular list."""

        def __init__(self, key, value, previous_element, next_element):
            """Class constructor."""
            self.key = key  # we also keep the key for O(1) access when removing the LRU.
            self.value = value
            self.previous = previous_element
            self.next = next_element

        def __str__(self):
            """Return string representation."""
            return '(%s, %s)' % (self.key, self.value)

    def __init__(self, max_size=DEFAULT_MAX_SIZE):
        """Class constructor."""
        self._data = {}
        self._lock = threading.Lock()
        self._max_size = max_size
        self._lru = None
        self._mru = None

    def test_and_set(self, key, value):
        """
        Set an item in the cache and return the previous value.

        :param key: object key
        :type args: object
        :param value: object value
        :type kwargs: object

        :return: previous value if any. None otherwise
        :rtype: object
        """
        with self._lock:
            node = self._data.get(key)
            to_return = node.value if node else None
            if node is None:
                node = SimpleLruCache._Node(key, value, None, None)
            node = self._bubble_up(node)
            self._data[key] = node
            self._rollover()
            return to_return

    def clear(self):
        """Clear the cache."""
        self._data = {}
        self._lru = None
        self._mru = None

    def _bubble_up(self, node):
        """Send node to the top of the list (mark it as the MRU)."""
        if node is None:
            return None

        # First item, just set lru & mru
        if not self._data:
            self._lru = node
            self._mru = node
            return node

        # MRU, just return it
        if node is self._mru:
            return node

        # LRU, update pointer and end-of-list
        if node is self._lru:
            self._lru = node.next
            self._lru.previous = None

        if node.previous is not None:
            node.previous.next = node.next
        if node.next is not None:
            node.next.previous = node.previous

        node.previous = self._mru
        node.previous.next = node
        node.next = None
        self._mru = node

        return node

    def _rollover(self):
        """Check we're within the size limit. Otherwise drop the LRU."""
        if len(self._data) > self._max_size:
            next_item = self._lru.next
            del self._data[self._lru.key]
            self._lru = next_item
            self._lru.previous = None

    def __str__(self):
        """User friendly representation of cache."""
        nodes = []
        node = self._mru
        while node is not None:
            nodes.append('\t<%s: %s>  -->' % (node.key, node.value))
            node = node.previous
        return '<MRU>\n' + '\n'.join(nodes) + '\n<LRU>'



"""SSE Notification definitions."""
import abc
import json
from enum import Enum

from future.utils import raise_from
from six import add_metaclass

from splitio.util.decorators import abstract_property
from splitio.util import utctime_ms
from splitio.push.sse import SSE_EVENT_ERROR, SSE_EVENT_MESSAGE


class EventType(Enum):
    """Event type enumeration."""

    MESSAGE = SSE_EVENT_MESSAGE
    ERROR = SSE_EVENT_ERROR


class MessageType(Enum):
    """Message type enumeration."""

    UPDATE = 0
    OCCUPANCY = 1
    CONTROL = 2


class UpdateType(Enum):
    """Message type enumeration."""

    SPLIT_UPDATE = 'SPLIT_UPDATE'
    SPLIT_KILL = 'SPLIT_KILL'
    SEGMENT_UPDATE = 'SEGMENT_UPDATE'


class ControlType(Enum):
    """Control type enumeration."""

    STREAMING_ENABLED = 'STREAMING_ENABLED'
    STREAMING_PAUSED = 'STREAMING_PAUSED'
    STREAMING_DISABLED = 'STREAMING_DISABLED'


TAG_OCCUPANCY = '[meta]occupancy'


class EventParsingException(Exception):
    """Exception to be raised on parser errors."""

    pass


@add_metaclass(abc.ABCMeta)  #pylint:disable=too-few-public-methods
class BaseEvent(object):
    """Base event that reqiures subclasses tu have a type."""

    @abstract_property
    def event_type(self):  #pylint:disable=no-self-use
        """
        Return the event type.

        :returns: The type of this parsed event.
        :rtype: EventType
        """
        pass


class AblyError(BaseEvent):
    """Ably Error message."""

    def __init__(self, code, status_code, message, href):
        """
        Class constructor.

        :param code: error code
        :type code: int

        :param status_code: http status cude
        :type status_code: int

        :param message: error message
        :type message: str

        :param href: link to error description
        :type href: str
        """
        self._code = code
        self._status_code = status_code
        self._message = message
        self._href = href
        self._timestamp = utctime_ms()

    @property
    def event_type(self):  #pylint:disable=no-self-use
        """
        Return the event type.

        :returns: The type of this parsed event.
        :rtype: MessageType
        """
        return EventType.ERROR

    @property
    def code(self):
        """
        Return the error code.

        :returns: ably error code.
        :rtype: int
        """
        return self._code

    @property
    def status_code(self):
        """
        Return the http status code.

        :returns: http status error code.
        :rtype: int
        """
        return self._status_code

    @property
    def message(self):
        """
        Return the ably error message.

        :returns: ably error message.
        :rtype: str
        """
        return self._message

    @property
    def href(self):
        """
        Return the link of the error description.

        :returns: error description url
        :rtype: str
        """
        return self._href

    @property
    def timestamp(self):
        """
        Return a the timestamp when this error was constructed.

        :returns: approximate error timestamp
        :rtype: int
        """
        return self._timestamp

    def should_be_ignored(self):
        """
        Return whether this error should be ignored or not.

        :returns: True if this error should be ignored. False otherwise.
        :rtype: bool
        """
        return self._code < 40000 or self._code > 49999


    def is_retryable(self):
        """
        Return whether this error is retryable or not.

        :returns: True if this error is retryable. False otherwise.
        :rtype: bool
        """
        return self._code >= 40140 and self._code <= 40149

    def __str__(self):
        """Return string representation."""
        return "AblyError - code=%d, status=%d, message=%s, href=%s" % \
            (self.code, self.status_code, self.message, self.href)


@add_metaclass(abc.ABCMeta)
class BaseMessage(BaseEvent):
    """Message type event."""

    def __init__(self, channel, timestamp):
        """
        Construct a message's base structure.

        :param channel: channel where the notification was received.
        :type channel: str
        """
        self._channel = channel
        self._timestamp = timestamp

    @property
    def channel(self):
        """
        Return the channel where the message arrived.

        :returns: channel
        :rtype: str
        """
        return self._channel

    @property
    def timestamp(self):
        """
        Return the timestamp when the message was sent.

        :returns: message sending timestamp
        :rtype: int
        """
        return self._timestamp

    @property
    def event_type(self):  #pylint:disable=no-self-use
        """
        Return the event type.

        :returns: The type of this parsed event.
        :rtype: MessageType
        """
        return EventType.MESSAGE

    @abstract_property
    def message_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Message.
        :rtype: MessageType
        """
        pass


class OccupancyMessage(BaseMessage):
    """Ably publisher occupancy notification."""

    def __init__(self, channel, timestamp, publishers):
        """
        Construct an occupancy message.

        :param channel: channel where occupancy is being announced.
        :type channel: str

        :param publishers: number of active publishers attached to this channel.
        :type data: int
        """
        BaseMessage.__init__(self, channel, timestamp)
        self._publishers = publishers

    @property
    def message_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Message.
        :rtype: MessageType
        """
        return MessageType.OCCUPANCY

    @property
    def channel(self):
        """
        Return the channel on which this message was received.

        :returns: channel name
        :rtype: str
        """
        return self._channel.replace('[?occupancy=metrics.publishers]', '')

    @property
    def publishers(self):
        """
        Return the number of publishers of this channel.

        :returns: attahed publisher count.
        :rtype: int
        """
        return self._publishers

    def __str__(self):
        """Return string representation."""
        return "Occupancy - channel=%s, publishers=%d" % (self.channel, self.publishers)


@add_metaclass(abc.ABCMeta)
class BaseUpdate(BaseMessage):
    """Split data update notification."""

    def __init__(self, channel, timestamp, change_number):
        """
        Construct an update event.

        :param data: raw message data.
        :type data: dict

        :param channel: channel where the message came from.
        :type channel: str
        """
        BaseMessage.__init__(self, channel, timestamp)
        self._change_number = change_number

    @abstract_property
    def update_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Update.
        :rtype: UpdateType
        """
        pass

    @property
    def message_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed event.
        :rtype: MessageType
        """
        return MessageType.UPDATE

    @property
    def change_number(self):
        """
        Return the change number associated with the data update.

        :returns: change number
        :rtype: int
        """
        return self._change_number


class SplitChangeUpdate(BaseUpdate):
    """Split Change notification."""

    def __init__(self, channel, timestamp, change_number):
        """Class constructor."""
        BaseUpdate.__init__(self, channel, timestamp, change_number)

    @property
    def update_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Update.
        :rtype: UpdateType
        """
        return UpdateType.SPLIT_UPDATE

    def __str__(self):
        """Return string representation."""
        return "SplitChange - changeNumber=%d" % (self.change_number)


class SplitKillUpdate(BaseUpdate):
    """Split Kill notification."""

    def __init__(self, channel, timestamp, change_number, split_name, default_treatment):  #pylint:disable=too-many-arguments
        """Class constructor."""
        BaseUpdate.__init__(self, channel, timestamp, change_number)
        self._split_name = split_name
        self._default_treatment = default_treatment

    @property
    def update_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Update.
        :rtype: UpdateType
        """
        return UpdateType.SPLIT_KILL

    @property
    def split_name(self):
        """
        Return the name of the killed split.

        :returns: name of the killed split
        :rtype: str
        """
        return self._split_name

    @property
    def default_treatment(self):
        """
        Return the default treatment.

        :returns: default treatment
        :rtype: str
        """
        return self._default_treatment

    def __str__(self):
        """Return string representation."""
        return "SplitKill - changeNumber=%d, name=%s, defaultTreatment=%s" % \
            (self.change_number, self.split_name, self.default_treatment)


class SegmentChangeUpdate(BaseUpdate):
    """Segment Change notification."""

    def __init__(self, channel, timestamp, change_number, segment_name):
        """Class constructor."""
        BaseUpdate.__init__(self, channel, timestamp, change_number)
        self._segment_name = segment_name

    @property
    def update_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed Update.
        :rtype: UpdateType
        """
        return UpdateType.SEGMENT_UPDATE

    @property
    def segment_name(self):
        """
        Return the semgent name associated with the data update.

        :returns: segment name
        :rtype: str
        """
        return self._segment_name

    def __str__(self):
        """Return string representation."""
        return "SegmentChange - changeNumber=%d, name=%s" % (self.change_number, self.segment_name)


class ControlMessage(BaseMessage):
    """Control notification."""

    def __init__(self, channel, timestamp, control_type):
        """Class constructor."""
        BaseMessage.__init__(self, channel, timestamp)
        self._control_type = ControlType(control_type)

    @property
    def message_type(self):  #pylint:disable=no-self-use
        """
        Return the message type.

        :returns: The type of this parsed event.
        :rtype: MessageType
        """
        return MessageType.CONTROL

    @property
    def control_type(self):
        """
        Return the associated control type.

        :returns: control type
        :rtype: ControlType
        """
        return self._control_type

    def __str__(self):
        """Return string representation."""
        return "Control - type=%s" % (self.control_type.name)


def _parse_update(channel, timestamp, data):
    """
    Parse a message of update type.

    :param channel: channel name
    :type data: str

    :param data: raw incoming event
    :type data: dict

    :returns: Parsed ably error notification.
    :rtype: BaseUpdate
    """
    update_type = UpdateType(data['type'])
    change_number = data['changeNumber']
    if update_type == UpdateType.SPLIT_UPDATE:
        return SplitChangeUpdate(channel, timestamp, change_number)
    elif update_type == UpdateType.SPLIT_KILL:
        return SplitKillUpdate(channel, timestamp, change_number,
                               data['splitName'], data['defaultTreatment'])
    elif update_type == UpdateType.SEGMENT_UPDATE:
        return SegmentChangeUpdate(channel, timestamp, change_number, data['segmentName'])
    raise EventParsingException('unrecognized event type %s' % update_type)


def _parse_message(data):
    """
    Parse a message event into a concrete class.

    :param data: raw incoming event.
    :type data: dict:

    :returns: Parsed ably error notification.
    :rtype: BaseEvent
    """
    if not all(k in data for k in ['data', 'channel']):
        return None
    channel = data['channel']
    timestamp = data['timestamp']
    parsed_data = json.loads(data['data'])
    if data.get('name') == TAG_OCCUPANCY:
        return OccupancyMessage(channel, timestamp, parsed_data['metrics']['publishers'])
    elif parsed_data['type'] == 'CONTROL':
        return ControlMessage(channel, timestamp, parsed_data['controlType'])
    elif parsed_data['type'] in UpdateType.__members__:
        return _parse_update(channel, timestamp, parsed_data)
    raise EventParsingException('unrecognized message type %s' % parsed_data['type'])


def _parse_error(data):
    """
    Parse an error message into a concrete class.

    :param data: raw incoming event.
    :type data: dict:

    :returns: Parsed ably error notification.
    :rtype: AblyError
    """
    return AblyError(data.get('code'), data.get('statusCode'),
                     data.get('message'), data.get('href'))


def parse_incoming_event(raw_event):
    """
    Parse a raw event as received by the sse client.

    :param raw_event: raw SSE Event
    :type raw_event: splitio.push.sse.SSEEvent

    :returns: an event parsed to it's concrete type.
    :rtype: BaseEvent
    """
    if raw_event is None:
        return None

    try:
        parsed_data = json.loads(raw_event.data)
    except Exception as exc:  #pylint:disable=broad-except
        raise_from(EventParsingException('Error parsing json'), exc)

    try:
        event_type = EventType(raw_event.event)
    except ValueError as exc:
        raise_from(exc, 'unknown event type %s' % raw_event.event)

    return {
        EventType.ERROR: _parse_error,
        EventType.MESSAGE: _parse_message,
    }[event_type](parsed_data)



"""In memory storage classes."""
from __future__ import absolute_import

import logging
import threading
from collections import Counter

from six.moves import queue
from splitio.models.segments import Segment
from splitio.storage import SplitStorage, SegmentStorage, ImpressionStorage, EventStorage, \
    TelemetryStorage

MAX_SIZE_BYTES = 5 * 1024 * 1024


_LOGGER = logging.getLogger(__name__)


class InMemorySplitStorage(SplitStorage):
    """InMemory implementation of a split storage."""

    def __init__(self):
        """Constructor."""
        self._lock = threading.RLock()
        self._splits = {}
        self._change_number = -1
        self._traffic_types = Counter()

    def get(self, split_name):
        """
        Retrieve a split.

        :param split_name: Name of the feature to fetch.
        :type split_name: str

        :rtype: splitio.models.splits.Split
        """
        with self._lock:
            return self._splits.get(split_name)

    def fetch_many(self, split_names):
        """
        Retrieve splits.

        :param split_names: Names of the features to fetch.
        :type split_name: list(str)

        :return: A dict with split objects parsed from queue.
        :rtype: dict(split_name, splitio.models.splits.Split)
        """
        return {split_name: self.get(split_name) for split_name in split_names}

    def put(self, split):
        """
        Store a split.

        :param split: Split object.
        :type split: splitio.models.split.Split
        """
        with self._lock:
            if split.name in self._splits:
                self._decrease_traffic_type_count(self._splits[split.name].traffic_type_name)
            self._splits[split.name] = split
            self._increase_traffic_type_count(split.traffic_type_name)

    def remove(self, split_name):
        """
        Remove a split from storage.

        :param split_name: Name of the feature to remove.
        :type split_name: str

        :return: True if the split was found and removed. False otherwise.
        :rtype: bool
        """
        with self._lock:
            split = self._splits.get(split_name)
            if not split:
                _LOGGER.warning("Tried to delete nonexistant split %s. Skipping", split_name)
                return False

            self._splits.pop(split_name)
            self._decrease_traffic_type_count(split.traffic_type_name)
            return True

    def get_change_number(self):
        """
        Retrieve latest split change number.

        :rtype: int
        """
        with self._lock:
            return self._change_number

    def set_change_number(self, new_change_number):
        """
        Set the latest change number.

        :param new_change_number: New change number.
        :type new_change_number: int
        """
        with self._lock:
            self._change_number = new_change_number

    def get_split_names(self):
        """
        Retrieve a list of all split names.

        :return: List of split names.
        :rtype: list(str)
        """
        with self._lock:
            return list(self._splits.keys())

    def get_all_splits(self):
        """
        Return all the splits.

        :return: List of all the splits.
        :rtype: list
        """
        with self._lock:
            return list(self._splits.values())

    def is_valid_traffic_type(self, traffic_type_name):
        """
        Return whether the traffic type exists in at least one split in cache.

        :param traffic_type_name: Traffic type to validate.
        :type traffic_type_name: str

        :return: True if the traffic type is valid. False otherwise.
        :rtype: bool
        """
        with self._lock:
            return traffic_type_name in self._traffic_types

    def kill_locally(self, split_name, default_treatment, change_number):
        """
        Local kill for split

        :param split_name: name of the split to perform kill
        :type split_name: str
        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        with self._lock:
            if self.get_change_number() > change_number:
                return
            split = self._splits.get(split_name)
            if not split:
                return
            split.local_kill(default_treatment, change_number)
            self.put(split)

    def _increase_traffic_type_count(self, traffic_type_name):
        """
        Increase by one the count for a specific traffic type name.

        :param traffic_type_name: Traffic type to increase the count.
        :type traffic_type_name: str
        """
        self._traffic_types.update([traffic_type_name])

    def _decrease_traffic_type_count(self, traffic_type_name):
        """
        Decrease by one the count for a specific traffic type name.

        :param traffic_type_name: Traffic type to decrease the count.
        :type traffic_type_name: str
        """
        self._traffic_types.subtract([traffic_type_name])
        self._traffic_types += Counter()


class InMemorySegmentStorage(SegmentStorage):
    """In-memory implementation of a segment storage."""

    def __init__(self):
        """Constructor."""
        self._segments = {}
        self._change_numbers = {}
        self._lock = threading.RLock()

    def get(self, segment_name):
        """
        Retrieve a segment.

        :param segment_name: Name of the segment to fetch.
        :type segment_name: str

        :rtype: str
        """
        with self._lock:
            fetched = self._segments.get(segment_name)
            if fetched is None:
                _LOGGER.warning(
                    "Tried to retrieve nonexistant segment %s. Skipping",
                    segment_name
                )
            return fetched

    def put(self, segment):
        """
        Store a segment.

        :param segment: Segment to store.
        :type segment: splitio.models.segment.Segment
        """
        with self._lock:
            self._segments[segment.name] = segment

    def update(self, segment_name, to_add, to_remove, change_number=None):
        """
        Update a split. Create it if it doesn't exist.

        :param segment_name: Name of the segment to update.
        :type segment_name: str
        :param to_add: Set of members to add to the segment.
        :type to_add: set
        :param to_remove: List of members to remove from the segment.
        :type to_remove: Set
        """
        with self._lock:
            if segment_name not in self._segments:
                self._segments[segment_name] = Segment(segment_name, to_add, change_number)
                return

            self._segments[segment_name].update(to_add, to_remove)
            if change_number is not None:
                self._segments[segment_name].change_number = change_number

    def get_change_number(self, segment_name):
        """
        Retrieve latest change number for a segment.

        :param segment_name: Name of the segment.
        :type segment_name: str

        :rtype: int
        """
        with self._lock:
            if segment_name not in self._segments:
                return None
            return self._segments[segment_name].change_number

    def set_change_number(self, segment_name, new_change_number):
        """
        Set the latest change number.

        :param segment_name: Name of the segment.
        :type segment_name: str
        :param new_change_number: New change number.
        :type new_change_number: int
        """
        with self._lock:
            if segment_name not in self._segments:
                return
            self._segments[segment_name].change_number = new_change_number

    def segment_contains(self, segment_name, key):
        """
        Check whether a specific key belongs to a segment in storage.

        :param segment_name: Name of the segment to search in.
        :type segment_name: str
        :param key: Key to search for.
        :type key: str

        :return: True if the segment contains the key. False otherwise.
        :rtype: bool
        """
        with self._lock:
            if segment_name not in self._segments:
                _LOGGER.warning(
                    "Tried to query members for nonexistant segment %s. Returning False",
                    segment_name
                )
                return False
            return self._segments[segment_name].contains(key)


class InMemoryImpressionStorage(ImpressionStorage):
    """In memory implementation of an impressions storage."""

    def __init__(self, queue_size):
        """
        Construct an instance.

        :param eventsQueueSize: How many events to queue before forcing a submission
        """
        self._impressions = queue.Queue(maxsize=queue_size)
        self._lock = threading.Lock()
        self._queue_full_hook = None

    def set_queue_full_hook(self, hook):
        """
        Set a hook to be called when the queue is full.

        :param h: Hook to be called when the queue is full
        """
        if callable(hook):
            self._queue_full_hook = hook

    def put(self, impressions):
        """
        Put one or more impressions in storage.

        :param impressions: List of one or more impressions to store.
        :type impressions: list
        """
        try:
            with self._lock:
                for impression in impressions:
                    self._impressions.put(impression, False)
            return True
        except queue.Full:
            if self._queue_full_hook is not None and callable(self._queue_full_hook):
                self._queue_full_hook()
            _LOGGER.warning(
                'Event queue is full, failing to add more events. \n'
                'Consider increasing parameter `eventQueueSize` in configuration'
            )
            return False

    def pop_many(self, count):
        """
        Pop the oldest N impressions from storage.

        :param count: Number of impressions to pop.
        :type count: int
        """
        impressions = []
        with self._lock:
            while not self._impressions.empty() and count > 0:
                impressions.append(self._impressions.get(False))
                count -= 1
        return impressions


class InMemoryEventStorage(EventStorage):
    """
    In memory storage for events.

    Supports adding and popping events.
    """

    def __init__(self, eventsQueueSize):
        """
        Construct an instance.

        :param eventsQueueSize: How many events to queue before forcing a submission
        """
        self._lock = threading.Lock()
        self._events = queue.Queue(maxsize=eventsQueueSize)
        self._queue_full_hook = None
        self._size = 0

    def set_queue_full_hook(self, hook):
        """
        Set a hook to be called when the queue is full.

        :param h: Hook to be called when the queue is full
        """
        if callable(hook):
            self._queue_full_hook = hook

    def put(self, events):
        """
        Add an event to storage.

        :param event: Event to be added in the storage
        """
        try:
            with self._lock:
                for event in events:
                    self._size += event.size

                    if self._size >= MAX_SIZE_BYTES:
                        self._queue_full_hook()
                        return False

                    self._events.put(event.event, False)
            return True
        except queue.Full:
            if self._queue_full_hook is not None and callable(self._queue_full_hook):
                self._queue_full_hook()
            _LOGGER.warning(
                'Events queue is full, failing to add more events. \n'
                'Consider increasing parameter `eventsQueueSize` in configuration'
            )
            return False

    def pop_many(self, count):
        """
        Pop multiple items from the storage.

        :param count: number of items to be retrieved and removed from the queue.
        """
        events = []
        with self._lock:
            while not self._events.empty() and count > 0:
                events.append(self._events.get(False))
                count -= 1
        self._size = 0
        return events


class InMemoryTelemetryStorage(TelemetryStorage):
    """In-Memory implementation of telemetry storage interface."""

    def __init__(self):
        """Constructor."""
        self._latencies = {}
        self._gauges = {}
        self._counters = {}
        self._latencies_lock = threading.Lock()
        self._gauges_lock = threading.Lock()
        self._counters_lock = threading.Lock()

    def inc_latency(self, name, bucket):
        """
        Add a latency.

        :param name: Name of the latency metric.
        :type name: str
        :param value: Value of the latency metric.
        :tyoe value: int
        """
        if not 0 <= bucket <= 21:
            _LOGGER.warning('Incorect bucket "%d" for latency "%s". Ignoring.', bucket, name)
            return

        with self._latencies_lock:
            latencies = self._latencies.get(name, [0] * 22)
            latencies[bucket] += 1
            self._latencies[name] = latencies

    def inc_counter(self, name):
        """
        Increment a counter.

        :param name: Name of the counter metric.
        :type name: str
        """
        with self._counters_lock:
            counter = self._counters.get(name, 0)
            counter += 1
            self._counters[name] = counter

    def put_gauge(self, name, value):
        """
        Add a gauge metric.

        :param name: Name of the gauge metric.
        :type name: str
        :param value: Value of the gauge metric.
        :type value: int
        """
        with self._gauges_lock:
            self._gauges[name] = value

    def pop_counters(self):
        """
        Get all the counters.

        :rtype: list
        """
        with self._counters_lock:
            try:
                return self._counters
            finally:
                self._counters = {}

    def pop_gauges(self):
        """
        Get all the gauges.

        :rtype: list

        """
        with self._gauges_lock:
            try:
                return self._gauges
            finally:
                self._gauges = {}

    def pop_latencies(self):
        """
        Get all latencies.

        :rtype: list
        """
        with self._latencies_lock:
            try:
                return self._latencies
            finally:
                self._latencies = {}



"""General purpose SDK utilities."""

import inspect
import socket
from collections import namedtuple
from splitio.version import __version__

SdkMetadata = namedtuple(
    'SdkMetadata',
    ['sdk_version', 'instance_name', 'instance_ip']
)


def _get_ip():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        sock.connect(('10.255.255.255', 1))
        ip_address = sock.getsockname()[0]
    except Exception:  # pylint: disable=broad-except
        ip_address = 'unknown'
    finally:
        sock.close()
    return ip_address


def _get_hostname(ip_address):
    return 'unknown' if ip_address == 'unknown' else 'ip-' + ip_address.replace('.', '-')


def _get_hostname_and_ip(config):
    if config.get('IPAddressesEnabled') is False:
        return 'NA', 'NA'
    ip_from_config = config.get('machineIp')
    machine_from_config = config.get('machineName')
    ip_address = ip_from_config if ip_from_config is not None else _get_ip()
    hostname = machine_from_config if machine_from_config is not None else _get_hostname(ip_address)
    return ip_address, hostname


def get_metadata(config):
    """
    Gather SDK metadata and return a tuple with such info.

    :param config: User supplied config augmented with defaults.
    :type config: dict

    :return: SDK Metadata information.
    :rtype: SdkMetadata
    """
    version = 'python-%s' % __version__
    ip_address, hostname = _get_hostname_and_ip(config)
    return SdkMetadata(version, hostname, ip_address)



import logging
from six.moves import queue

from splitio.api import APIException


_LOGGER = logging.getLogger(__name__)


class EventSynchronizer(object):
    def __init__(self, events_api, storage, bulk_size):
        """
        Class constructor.

        :param events_api: Events Api object to send data to the backend
        :type events_api: splitio.api.events.EventsAPI
        :param storage: Events Storage
        :type storage: splitio.storage.EventStorage
        :param bulk_size: How many events to send per push.
        :type bulk_size: int

        """
        self._api = events_api
        self._event_storage = storage
        self._bulk_size = bulk_size
        self._failed = queue.Queue()

    def _get_failed(self):
        """Return up to <BULK_SIZE> events stored in the failed eventes queue."""
        events = []
        count = 0
        while count < self._bulk_size:
            try:
                events.append(self._failed.get(False))
                count += 1
            except queue.Empty:
                # If no more items in queue, break the loop
                break
        return events

    def _add_to_failed_queue(self, events):
        """
        Add events that were about to be sent to a secondary queue for failed sends.

        :param events: List of events that failed to be pushed.
        :type events: list
        """
        for event in events:
            self._failed.put(event, False)

    def synchronize_events(self):
        """Send events from both the failed and new queues."""
        to_send = self._get_failed()
        if len(to_send) < self._bulk_size:
            # If the amount of previously failed items is less than the bulk
            # size, try to complete with new events from storage
            to_send.extend(self._event_storage.pop_many(self._bulk_size - len(to_send)))

        if not to_send:
            return

        try:
            self._api.flush_events(to_send)
        except APIException:
            _LOGGER.error('Exception raised while reporting events')
            _LOGGER.debug('Exception information: ', exc_info=True)
            self._add_to_failed_queue(to_send)



"""Synchronizer module."""

import abc
import logging
import threading

from six import add_metaclass
from future.utils import raise_from
from splitio.api import APIException


_LOGGER = logging.getLogger(__name__)


class SplitSynchronizers(object):
    """SplitSynchronizers."""

    def __init__(self, split_sync, segment_sync, impressions_sync, events_sync, telemetry_sync,  # pylint:disable=too-many-arguments
                 impressions_count_sync):
        """
        Class constructor.

        :param split_sync: sync for splits
        :type split_sync: splitio.sync.split.SplitSynchronizer
        :param segment_sync: sync for segments
        :type segment_sync: splitio.sync.segment.SegmentSynchronizer
        :param impressions_sync: sync for impressions
        :type impressions_sync: splitio.sync.impression.ImpressionSynchronizer
        :param events_sync: sync for events
        :type events_sync: splitio.sync.event.EventSynchronizer
        :param telemetry_sync: sync for telemetry
        :type telemetry_sync: splitio.sync.telemetry.TelemetrySynchronizer
        :param impressions_count_sync: sync for impression_counts
        :type impressions_count_sync: splitio.sync.impression.ImpressionsCountSynchronizer
        """
        self._split_sync = split_sync
        self._segment_sync = segment_sync
        self._impressions_sync = impressions_sync
        self._events_sync = events_sync
        self._telemetry_sync = telemetry_sync
        self._impressions_count_sync = impressions_count_sync

    @property
    def split_sync(self):
        """Return split synchonizer."""
        return self._split_sync

    @property
    def segment_sync(self):
        """Return segment synchonizer."""
        return self._segment_sync

    @property
    def impressions_sync(self):
        """Return impressions synchonizer."""
        return self._impressions_sync

    @property
    def events_sync(self):
        """Return events synchonizer."""
        return self._events_sync

    @property
    def telemetry_sync(self):
        """Return telemetry synchonizer."""
        return self._telemetry_sync

    @property
    def impressions_count_sync(self):
        """Return impressions count synchonizer."""
        return self._impressions_count_sync


class SplitTasks(object):
    """SplitTasks."""

    def __init__(self, split_task, segment_task, impressions_task, events_task, telemetry_task,  # pylint:disable=too-many-arguments
                 impressions_count_task):
        """
        Class constructor.

        :param split_task: sync for splits
        :type split_task: splitio.tasks.split_sync.SplitSynchronizationTask
        :param segment_task: sync for segments
        :type segment_task: splitio.tasks.segment_sync.SegmentSynchronizationTask
        :param impressions_task: sync for impressions
        :type impressions_task: splitio.tasks.impressions_sync.ImpressionsSyncTask
        :param events_task: sync for events
        :type events_task: splitio.tasks.events_sync.EventsSyncTask
        :param telemetry_task: sync for telemetry
        :type telemetry_task: splitio.tasks.telemetry_sync.TelemetrySynchronizationTask
        :param impressions_count_task: sync for impression_counts
        :type impressions_count_task: splitio.tasks.impressions_sync.ImpressionsCountSyncTask
        """
        self._split_task = split_task
        self._segment_task = segment_task
        self._impressions_task = impressions_task
        self._events_task = events_task
        self._telemetry_task = telemetry_task
        self._impressions_count_task = impressions_count_task

    @property
    def split_task(self):
        """Return split sync task."""
        return self._split_task

    @property
    def segment_task(self):
        """Return segment sync task."""
        return self._segment_task

    @property
    def impressions_task(self):
        """Return impressions sync task."""
        return self._impressions_task

    @property
    def events_task(self):
        """Return events sync task."""
        return self._events_task

    @property
    def telemetry_task(self):
        """Return telemetry sync task."""
        return self._telemetry_task

    @property
    def impressions_count_task(self):
        """Return impressions count sync task."""
        return self._impressions_count_task


@add_metaclass(abc.ABCMeta)
class BaseSynchronizer(object):
    """Synchronizer interface."""

    @abc.abstractmethod
    def synchronize_segment(self, segment_name, till):
        """
        Synchronize particular segment.

        :param segment_name: segment associated
        :type segment_name: str
        :param till: to fetch
        :type till: int
        """
        pass

    @abc.abstractmethod
    def synchronize_splits(self, till):
        """
        Synchronize all splits.

        :param till: to fetch
        :type till: int
        """
        pass

    @abc.abstractmethod
    def sync_all(self):
        """Synchronize all split data."""
        pass

    @abc.abstractmethod
    def start_periodic_fetching(self):
        """Start fetchers for splits and segments."""
        pass

    @abc.abstractmethod
    def stop_periodic_fetching(self):
        """Stop fetchers for splits and segments."""
        pass

    @abc.abstractmethod
    def start_periodic_data_recording(self):
        """Start recorders."""
        pass

    @abc.abstractmethod
    def stop_periodic_data_recording(self, blocking):
        """Stop recorders."""
        pass

    @abc.abstractmethod
    def kill_split(self, split_name, default_treatment, change_number):
        """
        Kill a split locally.

        :param split_name: name of the split to perform kill
        :type split_name: str
        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        pass

    @abc.abstractmethod
    def shutdown(self, blocking):
        """
        Stop tasks.

        :param blocking:flag to wait until tasks are stopped
        :type blocking: bool
        """
        pass


class Synchronizer(BaseSynchronizer):
    """Synchronizer."""

    def __init__(self, split_synchronizers, split_tasks):
        """
        Class constructor.

        :param split_synchronizers: syncs for performing synchronization of segments and splits
        :type split_synchronizers: splitio.sync.synchronizer.SplitSynchronizers
        :param split_tasks: tasks for starting/stopping tasks
        :type split_tasks: splitio.sync.synchronizer.SplitTasks
        """
        self._split_synchronizers = split_synchronizers
        self._split_tasks = split_tasks

    def _synchronize_segments(self):
        _LOGGER.debug('Starting segments synchronization')
        return self._split_synchronizers.segment_sync.synchronize_segments()

    def synchronize_segment(self, segment_name, till):
        """
        Synchronize particular segment.

        :param segment_name: segment associated
        :type segment_name: str
        :param till: to fetch
        :type till: int
        """
        _LOGGER.debug('Synchronizing segment %s', segment_name)
        return self._split_synchronizers.segment_sync.synchronize_segment(segment_name, till)

    def synchronize_splits(self, till):
        """
        Synchronize all splits.

        :param till: to fetch
        :type till: int
        """
        _LOGGER.debug('Starting splits synchronization')
        return self._split_synchronizers.split_sync.synchronize_splits(till)

    def sync_all(self):
        """Synchronize all split data."""
        try:
            self.synchronize_splits(None)
            if not self._synchronize_segments():
                _LOGGER.error('Failed syncing segments')
                raise RuntimeError('Failed syncing segments')
        except APIException as exc:
            _LOGGER.error('Failed syncing splits')
            raise_from(APIException('Failed to sync splits'), exc)

    def shutdown(self, blocking):
        """
        Stop tasks.

        :param blocking:flag to wait until tasks are stopped
        :type blocking: bool
        """
        _LOGGER.debug('Shutting down tasks.')
        self._split_synchronizers.segment_sync.shutdown()
        self.stop_periodic_fetching()
        self.stop_periodic_data_recording(blocking)

    def start_periodic_fetching(self):
        """Start fetchers for splits and segments."""
        _LOGGER.debug('Starting periodic data fetching')
        self._split_tasks.split_task.start()
        self._split_tasks.segment_task.start()

    def stop_periodic_fetching(self):
        """Stop fetchers for splits and segments."""
        _LOGGER.debug('Stopping periodic fetching')
        self._split_tasks.split_task.stop()
        self._split_tasks.segment_task.stop()

    def start_periodic_data_recording(self):
        """Start recorders."""
        _LOGGER.debug('Starting periodic data recording')
        self._split_tasks.impressions_task.start()
        self._split_tasks.events_task.start()
        self._split_tasks.telemetry_task.start()
        self._split_tasks.impressions_count_task.start()

    def stop_periodic_data_recording(self, blocking):
        """
        Stop recorders.

        :param blocking: flag to wait until tasks are stopped
        :type blocking: bool
        """
        _LOGGER.debug('Stopping periodic data recording')
        if blocking:
            events = []
            for task in [self._split_tasks.impressions_task,
                         self._split_tasks.events_task,
                         self._split_tasks.impressions_count_task]:
                stop_event = threading.Event()
                task.stop(stop_event)
                events.append(stop_event)
            if all(event.wait() for event in events):
                _LOGGER.debug('all tasks finished successfully.')
        else:
            self._split_tasks.impressions_task.stop()
            self._split_tasks.events_task.stop()
            self._split_tasks.impressions_count_task.stop()
        self._split_tasks.telemetry_task.stop()

    def kill_split(self, split_name, default_treatment, change_number):
        """
        Kill a split locally.

        :param split_name: name of the split to perform kill
        :type split_name: str
        :param default_treatment: name of the default treatment to return
        :type default_treatment: str
        :param change_number: change_number
        :type change_number: int
        """
        self._split_synchronizers.split_sync.kill_split(split_name, default_treatment,
                                                        change_number)


class LocalhostSynchronizer(BaseSynchronizer):
    """LocalhostSynchronizer."""

    def __init__(self, split_synchronizers, split_tasks):
        """
        Class constructor.

        :param split_synchronizers: syncs for performing synchronization of segments and splits
        :type split_synchronizers: splitio.sync.synchronizer.SplitSynchronizers
        :param split_tasks: tasks for starting/stopping tasks
        :type split_tasks: splitio.sync.synchronizer.SplitTasks
        """
        self._split_synchronizers = split_synchronizers
        self._split_tasks = split_tasks

    def sync_all(self):
        """Synchronize all split data."""
        try:
            self._split_synchronizers.split_sync.synchronize_splits(None)
        except APIException as exc:
            _LOGGER.error('Failed syncing splits')
            raise_from(APIException('Failed to sync splits'), exc)

    def start_periodic_fetching(self):
        """Start fetchers for splits and segments."""
        _LOGGER.debug('Starting periodic data fetching')
        self._split_tasks.split_task.start()

    def stop_periodic_fetching(self):
        """Stop fetchers for splits and segments."""
        _LOGGER.debug('Stopping periodic fetching')
        self._split_tasks.split_task.stop()

    def kill_split(self, split_name, default_treatment, change_number):
        """Kill a split locally."""
        raise NotImplementedError()

    def synchronize_splits(self, till):
        """Synchronize all splits."""
        raise NotImplementedError()

    def synchronize_segment(self, segment_name, till):
        """Synchronize particular segment."""
        raise NotImplementedError()

    def start_periodic_data_recording(self):
        """Start recorders."""
        pass

    def stop_periodic_data_recording(self, blocking):
        """Stop recorders."""
        pass

    def shutdown(self, blocking):
        """
        Stop tasks.

        :param blocking:flag to wait until tasks are stopped
        :type blocking: bool
        """
        self.stop_periodic_fetching()



"""Split conditions module."""

from enum import Enum
from future.utils import python_2_unicode_compatible
import six

from splitio.models.grammar import matchers
from splitio.models.grammar import partitions

_MATCHER_COMBINERS = {
    'AND': lambda ms, k, a, c: all(m.evaluate(k, a, c) for m in ms)
}


class ConditionType(Enum):
    """Split possible condition types."""

    WHITELIST = 'WHITELIST'
    ROLLOUT = 'ROLLOUT'


class Condition(object):
    """Condition object class."""

    def __init__(  #pylint: disable=too-many-arguments
            self,
            matcher_list,
            combiner, parts, label,
            condition_type=ConditionType.WHITELIST
    ):
        """
        Class constructor.

        :param matcher: A combining matcher
        :type matcher: CombiningMatcher
        :param parts: A list of partitions
        :type parts: list
        """
        self._matchers = matcher_list
        self._combiner = combiner
        self._partitions = tuple(parts)
        self._label = label
        self._condition_type = condition_type

    @property
    def matchers(self):
        """Return the list of matchers associated to the condition."""
        return self._matchers

    @property
    def partitions(self):
        """Return the list of partitions associated with the condition."""
        return self._partitions

    @property
    def label(self):
        """Return the label of this condition."""
        return self._label

    @property
    def condition_type(self):
        """Return the condition type."""
        return self._condition_type

    def matches(self, key, attributes=None, context=None):
        """
        Check whether the condition matches against user submitted input.

        :param key: User key
        :type key: splitio.client.key.Key
        :param attributes: User custom attributes.
        :type attributes: dict
        :param context: Evaluation context
        :type context: dict
        """
        return self._combiner(self._matchers, key, attributes, context)

    def get_segment_names(self):
        """
        Fetch segment names for all IN_SEGMENT matchers.

        :return: List of segment names
        :rtype: list(str)
        """
        return [
            matcher._segment_name for matcher in self.matchers  #pylint: disable=protected-access
            if isinstance(matcher, matchers.UserDefinedSegmentMatcher)
        ]

    @python_2_unicode_compatible
    def __str__(self):
        """Return the string representation of the condition."""
        return '{matcher} then split {parts}'.format(
            matcher=self._matchers, parts=','.join(
                '{size}:{treatment}'.format(size=partition.size,
                                            treatment=partition.treatment)
                for partition in self._partitions))

    def to_json(self):
        """Return the JSON representation of this condition."""
        return {
            'conditionType': self._condition_type.name,
            'label': self._label,
            'matcherGroup': {
                'combiner': next(
                    (k, v) for k, v in six.iteritems(_MATCHER_COMBINERS) if v == self._combiner
                )[0],
                'matchers': [m.to_json() for m in self.matchers]
            },
            'partitions': [p.to_json() for p in self.partitions]
        }


def from_raw(raw_condition):
    """
    Parse a condition from a JSON portion of splitChanges.

    :param raw_condition: JSON object extracted from a split's conditions array.
    :type raw_condition: dict

    :return: A condition object.
    :rtype: Condition
    """
    parsed_partitions = [
        partitions.from_raw(raw_partition)
        for raw_partition in raw_condition['partitions']
    ]

    matcher_objects = [matchers.from_raw(x) for x in raw_condition['matcherGroup']['matchers']]
    combiner = _MATCHER_COMBINERS[raw_condition['matcherGroup']['combiner']]
    label = raw_condition.get('label')

    condition_type = ConditionType(raw_condition.get('conditionType', ConditionType.WHITELIST))

    return Condition(matcher_objects, combiner, parsed_partitions, label, condition_type)



"""Threading utilities."""
from inspect import isclass
import threading


# python2 workaround
_EventClass = threading.Event if isclass(threading.Event) else threading._Event  #pylint:disable=protected-access,invalid-name


class EventGroup(object):
    """EventGroup that can be waited with an OR condition."""

    class Event(_EventClass):  #pylint:disable=too-few-public-methods
        """Threading event meant to be used in an group."""

        def __init__(self, shared_condition):
            """
            Construct an event.

            :param shared_condition: shared condition varaible.
            :type shared_condition: threading.Condition
            """
            _EventClass.__init__(self)
            self._shared_cond = shared_condition

        def set(self):
            """Set the event."""
            _EventClass.set(self)
            with self._shared_cond:
                self._shared_cond.notify()

    def __init__(self):
        """Construct an event group."""
        self._cond = threading.Condition()

    def make_event(self):
        """
        Make a new event associated to this waitable group.

        :returns: an event that can be awaited as part of a group
        :rtype: EventGroup.Event
        """
        return EventGroup.Event(self._cond)

    def wait(self, timeout=None):
        """
        Wait until one of the events is triggered.

        :param timeout: how many seconds to wait. None means forever.
        :type timeout: int

        :returns: True if the condition was notified within the specified timeout. False otherwise.
        :rtype: bool
        """
        with self._cond:
            return self._cond.wait(timeout)



import logging
from six.moves import queue

from splitio.api import APIException


_LOGGER = logging.getLogger(__name__)


class TelemetrySynchronizer(object):
    def __init__(self, api, storage):
        """
        Class constructor.

        :param api: Telemetry API Client.
        :type api: splitio.api.telemetry.TelemetryAPI
        :param storage: Telemetry Storage.
        :type storage: splitio.storage.InMemoryTelemetryStorage

        """
        self._api = api
        self._storage = storage

    def synchronize_telemetry(self):
        """
        Send latencies, counters and gauges to split BE.

        :return: True if synchronization is complete.
        :rtype: bool
        """
        try:
            latencies = self._storage.pop_latencies()
            if latencies:
                self._api.flush_latencies(latencies)
        except APIException:
            _LOGGER.error('Failed send telemetry/latencies to split BE.')
            _LOGGER.debug('Exception information: ', exc_info=True)

        try:
            counters = self._storage.pop_counters()
            if counters:
                self._api.flush_counters(counters)
        except APIException:
            _LOGGER.error('Failed send telemetry/counters to split BE.')
            _LOGGER.debug('Exception information: ', exc_info=True)

        try:
            gauges = self._storage.pop_gauges()
            if gauges:
                self._api.flush_gauges(gauges)
        except APIException:
            _LOGGER.error('Failed send telemetry/gauges to split BE.')
            _LOGGER.debug('Exception information: ', exc_info=True)



"""Split changes processing worker."""
import logging
import threading


_LOGGER = logging.getLogger(__name__)


class SplitWorker(object):
    """Split Worker for processing updates."""

    _centinel = object()

    def __init__(self, synchronize_split, split_queue):
        """
        Class constructor.

        :param synchronize_split: handler to perform split synchronization on incoming event
        :type synchronize_split: callable

        :param split_queue: queue with split updates notifications
        :type split_queue: queue
        """
        self._split_queue = split_queue
        self._handler = synchronize_split
        self._running = False
        self._worker = None

    def is_running(self):
        """Return whether the working is running."""
        return self._running

    def _run(self):
        """Run worker handler."""
        while self.is_running():
            event = self._split_queue.get()
            if not self.is_running():
                break
            if event == self._centinel:
                continue
            _LOGGER.debug('Processing split_update %d', event.change_number)
            try:
                self._handler(event.change_number)
            except Exception:  # pylint: disable=broad-except
                _LOGGER.error('Exception raised in split synchronization')
                _LOGGER.debug('Exception information: ', exc_info=True)

    def start(self):
        """Start worker."""
        if self.is_running():
            _LOGGER.debug('Worker is already running')
            return
        self._running = True

        _LOGGER.debug('Starting Split Worker')
        self._worker = threading.Thread(target=self._run, name='PushSplitWorker')
        self._worker.setDaemon(True)
        self._worker.start()

    def stop(self):
        """Stop worker."""
        _LOGGER.debug('Stopping Split Worker')
        if not self.is_running():
            _LOGGER.debug('Worker is not running')
            return
        self._running = False
        self._split_queue.put(self._centinel)



"""Worker pool module."""

import logging
from threading import Thread, Event
from six.moves import queue


_LOGGER = logging.getLogger(__name__)


class WorkerPool(object):
    """Worker pool class to implement single producer/multiple consumer."""

    def __init__(self, worker_count, worker_func):
        """
        Class constructor.

        :param worker_count: Number of workers for the pool.
        :type worker_func: Function to be executed by the workers whenever a messages is fetched.
        """
        self._failed = False
        self._incoming = queue.Queue()
        self._should_be_working = [True for _ in range(0, worker_count)]
        self._worker_events = [Event() for _ in range(0, worker_count)]
        self._threads = [
            Thread(target=self._wrapper, args=(i, worker_func), name="pool_worker_%d" % i)
            for i in range(0, worker_count)
        ]
        for thread in self._threads:
            thread.setDaemon(True)

    def start(self):
        """Start the workers."""
        for thread in self._threads:
            thread.start()

    @staticmethod
    def _safe_run(func, message):
        """
        Execute the user funcion for a given message without raising exceptions.

        :param func: User defined function.
        :type func: callable
        :param message: Message fetched from the queue.
        :param message: object

        :return True if no everything goes well. False otherwise.
        :rtype bool
        """
        try:
            func(message)
            return True
        except Exception:  # pylint: disable=broad-except
            _LOGGER.error("Something went wrong when processing message %s", message)
            _LOGGER.debug('Original traceback: ', exc_info=True)
            return False

    def _wrapper(self, worker_number, func):
        """
        Fetch message, execute tasks, and acknowledge results.

        :param worker_number: # (id) of worker whose function will be executed.
        :type worker_number: int
        :param func: User defined function.
        :type func: callable.
        """
        while self._should_be_working[worker_number]:
            try:
                message = self._incoming.get(True, 0.5)

                # For some reason message can be None in python2 implementation of queue.
                # This method must be both ignored and acknowledged with .task_done()
                # otherwise .join() will halt.
                if message is None:
                    _LOGGER.debug('spurious message received. acking and ignoring.')
                    self._incoming.task_done()
                    continue

                # If the task is successfully executed, the ack is done AFTERWARDS,
                # to avoid race conditions on SDK initialization.
                _LOGGER.debug("processing message '%s'", message)
                ok = self._safe_run(func, message)  # pylint: disable=invalid-name
                if not ok:
                    self._failed = True
                    _LOGGER.error(
                        ("Something went wrong during the execution, "
                         "removing message \"%s\" from queue."),
                        message
                    )
                self._incoming.task_done()
            except queue.Empty:
                # No message was fetched, just keep waiting.
                pass

        # Set my flag indicating that i have finished
        self._worker_events[worker_number].set()

    def submit_work(self, message):
        """
        Add a new message to the work-queue.

        :param message: New message to add.
        :type message: object.
        """
        self._incoming.put(message)
        _LOGGER.debug('queued message %s for processing.', message)

    def wait_for_completion(self):
        """Block until the work queue is empty."""
        _LOGGER.debug('waiting for all messages to be processed.')
        self._incoming.join()
        _LOGGER.debug('all messages processed.')
        old = self._failed
        self._failed = False
        return old

    def stop(self, event=None):
        """Stop all worker nodes."""
        async_stop = Thread(target=self._wait_workers_shutdown, args=(event,))
        async_stop.setDaemon(True)
        async_stop.start()

    def _wait_workers_shutdown(self, event):
        """
        Wait until all workers have finished, and set the event.

        :param event: Event to set as soon as all the workers have shut down.
        :type event: threading.Event
        """
        self.wait_for_completion()
        for index, _ in enumerate(self._should_be_working):
            self._should_be_working[index] = False

        if event is not None:
            for worker_event in self._worker_events:
                worker_event.wait()
            event.set()



"""Low-level SSE Client."""
import logging
import socket
import sys
from collections import namedtuple

try:  # try to import python3 names. fallback to python2
    from http.client import HTTPConnection, HTTPSConnection
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
    from httplib import HTTPConnection, HTTPSConnection


_LOGGER = logging.getLogger(__name__)


SSE_EVENT_ERROR = 'error'
SSE_EVENT_MESSAGE = 'message'


SSEEvent = namedtuple('SSEEvent', ['event_id', 'event', 'retry', 'data'])


__ENDING_CHARS = set(['\n', ''])
def __httpresponse_readline_py2(response):
    """
    Hacky `readline` implementation to be used with chunked transfers in python2.

    This makes syscalls in a loop, so not particularly efficient. Migrate to py3 now!

    :param response: HTTPConnection's response after a .request() call
    :type response: httplib.HTTPResponse

    :returns: a string with the read line
    :rtype: str
    """
    buf = []
    while True:
        read = response.read(1)
        buf.append(read)
        if read in __ENDING_CHARS:
            break

    return ''.join(buf)


_http_response_readline = (__httpresponse_readline_py2 if sys.version_info.major <= 2  #pylint:disable=invalid-name
                           else lambda response: response.readline())


class EventBuilder(object):
    """Event builder class."""

    _SEPARATOR = b':'

    def __init__(self):
        """Construct a builder."""
        self._lines = {}

    def process_line(self, line):
        """
        Process a new line.

        :param line: Line to process
        :type line: bytes
        """
        try:
            key, val = line.split(self._SEPARATOR, 1)
            self._lines[key.decode('utf8').strip()] = val.decode('utf8').strip()
        except ValueError:  # key without a value
            self._lines[line.decode('utf8').strip()] = None

    def build(self):
        """Construct an event with relevant fields."""
        return SSEEvent(self._lines.get('id'), self._lines.get('event'),
                        self._lines.get('retry'), self._lines.get('data'))


class SSEClient(object):
    """SSE Client implementation."""

    _DEFAULT_HEADERS = {'accept': 'text/event-stream'}
    _EVENT_SEPARATORS = set([b'\n', b'\r\n'])

    def __init__(self, callback):
        """
        Construct an SSE client.

        :param callback: function to call when an event is received
        :type callback: callable
        """
        self._conn = None
        self._event_callback = callback
        self._shutdown_requested = False

    def _read_events(self):
        """
        Read events from the supplied connection.

        :returns: True if the connection was ended by us. False if it was closed by the serve.
        :rtype: bool
        """
        try:
            response = self._conn.getresponse()
            event_builder = EventBuilder()
            while True:
                line = _http_response_readline(response)
                if line is None or len(line) <= 0:  # connection ended
                    break
                elif line.startswith(b':'):  # comment. Skip
                    _LOGGER.debug("skipping sse comment")
                    continue
                elif line in self._EVENT_SEPARATORS:
                    event = event_builder.build()
                    _LOGGER.debug("dispatching event: %s", event)
                    self._event_callback(event)
                    event_builder = EventBuilder()
                else:
                    event_builder.process_line(line)
        except Exception:  #pylint:disable=broad-except
            _LOGGER.debug('sse connection ended.')
            _LOGGER.debug('stack trace: ', exc_info=True)
        finally:
            self._conn.close()
            self._conn = None  # clear so it can be started again

        return self._shutdown_requested

    def start(self, url, extra_headers=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):  #pylint:disable=protected-access
        """
        Connect and start listening for events.

        :param url: url to connect to
        :type url: str

        :param extra_headers: additional headers
        :type extra_headers: dict[str, str]

        :param timeout: connection & read timeout
        :type timeout: float

        :returns: True if the connection was ended by us. False if it was closed by the serve.
        :rtype: bool
        """
        if self._conn is not None:
            raise RuntimeError('Client already started.')

        self._shutdown_requested = False
        url = urlparse(url)
        headers = self._DEFAULT_HEADERS.copy()
        headers.update(extra_headers if extra_headers is not None else {})
        self._conn = (HTTPSConnection(url.hostname, url.port, timeout=timeout)
                      if url.scheme == 'https'
                      else HTTPConnection(url.hostname, port=url.port, timeout=timeout))

        self._conn.request('GET', '%s?%s' % (url.path, url.query), headers=headers)
        return self._read_events()

    def shutdown(self):
        """Shutdown the current connection."""
        if self._conn is None:
            _LOGGER.warn("no sse connection has been started on this SSEClient instance. Ignoring")
            return

        if self._shutdown_requested:
            _LOGGER.warn("shutdown already requested")
            return

        self._shutdown_requested = True
        self._conn.sock.shutdown(socket.SHUT_RDWR)



"""Misc helpers for testing purposes."""


class Any(object):  #pylint:disable=too-few-public-methods
    """Crap that matches anything."""

    def __eq__(self, other):
        """Match anything."""
        return True



"""Synchronous HTTP Client for split API."""
from __future__ import division

from collections import namedtuple

from future.utils import raise_from
import requests

HttpResponse = namedtuple('HttpResponse', ['status_code', 'body'])


class HttpClientException(Exception):
    """HTTP Client exception."""

    def __init__(self, message):
        """
        Class constructor.

        :param message: Information on why this exception happened.
        :type message: str
        """
        Exception.__init__(self, message)


class HttpClient(object):
    """HttpClient wrapper."""

    SDK_URL = 'https://sdk.split.io/api'
    EVENTS_URL = 'https://events.split.io/api'
    AUTH_URL = 'https://auth.split.io/api'

    def __init__(self, timeout=None, sdk_url=None, events_url=None, auth_url=None):
        """
        Class constructor.

        :param timeout: How many milliseconds to wait until the server responds.
        :type timeout: int
        :param sdk_url: Optional alternative sdk URL.
        :type sdk_url: str
        :param events_url: Optional alternative events URL.
        :type events_url: str
        :param auth_url: Optional alternative auth URL.
        :type auth_url: str
        """
        self._timeout = timeout/1000 if timeout else None  # Convert ms to seconds.
        self._urls = {
            'sdk': sdk_url if sdk_url is not None else self.SDK_URL,
            'events': events_url if events_url is not None else self.EVENTS_URL,
            'auth': auth_url if auth_url is not None else self.AUTH_URL,
        }

    def _build_url(self, server, path):
        """
        Build URL according to server specified.

        :param server: Server for whith the request is being made.
        :type server: str
        :param path: URL path to be appended to base host.
        :type path: str

        :return: A fully qualified URL.
        :rtype: str
        """
        return self._urls[server] + path

    @staticmethod
    def _build_basic_headers(apikey):
        """
        Build basic headers with auth.

        :param apikey: API token used to identify backend calls.
        :type apikey: str
        """
        return {
            'Content-Type': 'application/json',
            'Authorization': "Bearer %s" % apikey
        }

    def get(self, server, path, apikey, query=None, extra_headers=None):  # pylint: disable=too-many-arguments
        """
        Issue a get request.

        :param server: Whether the request is for SDK server, Events server or Auth server.
        :typee server: str
        :param path: path to append to the host url.
        :type path: str
        :param apikey: api token.
        :type apikey: str
        :param query: Query string passed as dictionary.
        :type query: dict
        :param extra_headers: key/value pairs of possible extra headers.
        :type extra_headers: dict

        :return: Tuple of status_code & response text
        :rtype: HttpResponse
        """
        headers = self._build_basic_headers(apikey)
        if extra_headers is not None:
            headers.update(extra_headers)

        try:
            response = requests.get(
                self._build_url(server, path),
                params=query,
                headers=headers,
                timeout=self._timeout
            )
            return HttpResponse(response.status_code, response.text)
        except Exception as exc:  # pylint: disable=broad-except
            raise_from(HttpClientException('requests library is throwing exceptions'), exc)

    def post(self, server, path, apikey, body, query=None, extra_headers=None):  # pylint: disable=too-many-arguments
        """
        Issue a POST request.

        :param server: Whether the request is for SDK server or Events server.
        :typee server: str
        :param path: path to append to the host url.
        :type path: str
        :param apikey: api token.
        :type apikey: str
        :param body: body sent in the request.
        :type body: str
        :param query: Query string passed as dictionary.
        :type query: dict
        :param extra_headers: key/value pairs of possible extra headers.
        :type extra_headers: dict

        :return: Tuple of status_code & response text
        :rtype: HttpResponse
        """
        headers = self._build_basic_headers(apikey)

        if extra_headers is not None:
            headers.update(extra_headers)

        try:
            response = requests.post(
                self._build_url(server, path),
                json=body,
                params=query,
                headers=headers,
                timeout=self._timeout
            )
            return HttpResponse(response.status_code, response.text)
        except Exception as exc:  # pylint: disable=broad-except
            raise_from(HttpClientException('requests library is throwing exceptions'), exc)



"""Split Worker tests."""

import threading
import time
import pytest

from splitio.api import APIException
from splitio.tasks import split_sync
from splitio.storage import SplitStorage
from splitio.models.splits import Split
from splitio.sync.split import SplitSynchronizer


class SplitsSynchronizerTests(object):
    """Split synchronizer test cases."""

    def test_synchronize_splits_error(self, mocker):
        """Test that if fetching splits fails at some_point, the task will continue running."""
        storage = mocker.Mock(spec=SplitStorage)
        api = mocker.Mock()

        def run(x):
            raise APIException("something broke")
        run._calls = 0
        api.fetch_splits.side_effect = run
        storage.get_change_number.return_value = -1

        split_synchronizer = SplitSynchronizer(api, storage)

        with pytest.raises(APIException):
            split_synchronizer.synchronize_splits(1)

    def test_synchronize_splits(self, mocker):
        """Test split sync."""
        storage = mocker.Mock(spec=SplitStorage)

        def change_number_mock():
            change_number_mock._calls += 1
            if change_number_mock._calls == 1:
                return -1
            return 123
        change_number_mock._calls = 0
        storage.get_change_number.side_effect = change_number_mock

        api = mocker.Mock()
        splits = [{
           'changeNumber': 123,
           'trafficTypeName': 'user',
           'name': 'some_name',
           'trafficAllocation': 100,
           'trafficAllocationSeed': 123456,
           'seed': 321654,
           'status': 'ACTIVE',
           'killed': False,
           'defaultTreatment': 'off',
           'algo': 2,
           'conditions': [
               {
                   'partitions': [
                       {'treatment': 'on', 'size': 50},
                       {'treatment': 'off', 'size': 50}
                   ],
                   'contitionType': 'WHITELIST',
                   'label': 'some_label',
                   'matcherGroup': {
                       'matchers': [
                           {
                               'matcherType': 'WHITELIST',
                               'whitelistMatcherData': {
                                   'whitelist': ['k1', 'k2', 'k3']
                               },
                               'negate': False,
                           }
                       ],
                       'combiner': 'AND'
                   }
               }
            ]
        }]

        def get_changes(*args, **kwargs):
            get_changes.called += 1

            if get_changes.called == 1:
                return {
                    'splits': splits,
                    'since': -1,
                    'till': 123
                }
            else:
                return {
                    'splits': [],
                    'since': 123,
                    'till': 123
                }
        get_changes.called = 0

        api.fetch_splits.side_effect = get_changes
        split_synchronizer = SplitSynchronizer(api, storage)
        split_synchronizer.synchronize_splits()

        assert mocker.call(-1) in api.fetch_splits.mock_calls
        assert mocker.call(123) in api.fetch_splits.mock_calls

        inserted_split = storage.put.mock_calls[0][1][0]
        assert isinstance(inserted_split, Split)
        assert inserted_split.name == 'some_name'

    def test_not_called_on_till(self, mocker):
        """Test that sync is not called when till is less than previous changenumber"""
        storage = mocker.Mock(spec=SplitStorage)

        def change_number_mock():
            return 2
        storage.get_change_number.side_effect = change_number_mock

        def get_changes(*args, **kwargs):
            get_changes.called += 1
            return None

        get_changes.called = 0

        api = mocker.Mock()
        api.fetch_splits.side_effect = get_changes

        split_synchronizer = SplitSynchronizer(api, storage)
        split_synchronizer.synchronize_splits(1)

        assert get_changes.called == 0



"""SSEClient unit tests."""

import time
import threading
import pytest
from splitio.push.sse import SSEClient, SSEEvent
from tests.helpers.mockserver import SSEMockServer


class SSEClientTests(object):
    """SSEClient test cases."""

    def test_sse_client_disconnects(self):
        """Test correct initialization. Client ends the connection."""
        server = SSEMockServer()
        server.start()

        events = []
        def callback(event):
            """Callback."""
            events.append(event)

        client = SSEClient(callback)

        def runner():
            """SSE client runner thread."""
            assert client.start('http://127.0.0.1:' + str(server.port()))
        client_task = threading.Thread(target=runner)
        client_task.setDaemon(True)
        client_task.setName('client')
        client_task.start()
        with pytest.raises(RuntimeError):
            client_task.start()

        server.publish({'id': '1'})
        server.publish({'id': '2', 'event': 'message', 'data': 'abc'})
        server.publish({'id': '3', 'event': 'message', 'data': 'def'})
        server.publish({'id': '4', 'event': 'message', 'data': 'ghi'})
        time.sleep(1)
        client.shutdown()
        time.sleep(1)

        assert events == [
            SSEEvent('1', None, None, None),
            SSEEvent('2', 'message', None, 'abc'),
            SSEEvent('3', 'message', None, 'def'),
            SSEEvent('4', 'message', None, 'ghi')
        ]

        assert client._conn is None
        server.publish(server.GRACEFUL_REQUEST_END)
        server.stop()

    def test_sse_server_disconnects(self):
        """Test correct initialization. Server ends connection."""
        server = SSEMockServer()
        server.start()

        events = []
        def callback(event):
            """Callback."""
            events.append(event)

        client = SSEClient(callback)

        def runner():
            """SSE client runner thread."""
            assert client.start('http://127.0.0.1:' + str(server.port()))
        client_task = threading.Thread(target=runner)
        client_task.setDaemon(True)
        client_task.setName('client')
        client_task.start()

        server.publish({'id': '1'})
        server.publish({'id': '2', 'event': 'message', 'data': 'abc'})
        server.publish({'id': '3', 'event': 'message', 'data': 'def'})
        server.publish({'id': '4', 'event': 'message', 'data': 'ghi'})
        time.sleep(1)
        server.publish(server.GRACEFUL_REQUEST_END)
        server.stop()
        time.sleep(1)

        assert events == [
            SSEEvent('1', None, None, None),
            SSEEvent('2', 'message', None, 'abc'),
            SSEEvent('3', 'message', None, 'def'),
            SSEEvent('4', 'message', None, 'ghi')
        ]

        assert client._conn is None

    def test_sse_server_disconnects_abruptly(self):
        """Test correct initialization. Server ends connection."""
        server = SSEMockServer()
        server.start()
 
        events = []
        def callback(event):
            """Callback."""
            events.append(event)

        client = SSEClient(callback)

        def runner():
            """SSE client runner thread."""
            assert client.start('http://127.0.0.1:' + str(server.port()))
        client_task = threading.Thread(target=runner)
        client_task.setDaemon(True)
        client_task.setName('client')
        client_task.start()

        server.publish({'id': '1'})
        server.publish({'id': '2', 'event': 'message', 'data': 'abc'})
        server.publish({'id': '3', 'event': 'message', 'data': 'def'})
        server.publish({'id': '4', 'event': 'message', 'data': 'ghi'})
        time.sleep(1)
        server.publish(server.VIOLENT_REQUEST_END)
        server.stop()
        time.sleep(1)

        assert events == [
            SSEEvent('1', None, None, None),
            SSEEvent('2', 'message', None, 'abc'),
            SSEEvent('3', 'message', None, 'def'),
            SSEEvent('4', 'message', None, 'ghi')
        ]

        assert client._conn is None



"""Split Synchronization task."""

import logging
from splitio.tasks import BaseSynchronizationTask
from splitio.tasks.util.asynctask import AsyncTask


_LOGGER = logging.getLogger(__name__)


class SplitSynchronizationTask(BaseSynchronizationTask):
    """Split Synchronization task class."""
    def __init__(self, synchronize_splits, period):
        """
        Class constructor.

        :param synchronize_splits: Handler
        :type synchronize_splits: func
        :param period: Period of task
        :type period: int
        """
        self._period = period
        self._task = AsyncTask(synchronize_splits, period, on_init=None)

    def start(self):
        """Start the task."""
        self._task.start()

    def stop(self, event=None):
        """Stop the task. Accept an optional event to set when the task has finished."""
        self._task.stop(event)

    def is_running(self):
        """
        Return whether the task is running.

        :return: True if the task is running. False otherwise.
        :rtype bool
        """
        return self._task.running()



"""Push subsystem manager class and helpers."""

import logging
from threading import Timer

from splitio.api import APIException
from splitio.push.splitsse import SplitSSEClient
from splitio.push.parser import parse_incoming_event, EventParsingException, EventType, \
    MessageType
from splitio.push.processor import MessageProcessor
from splitio.push.status_tracker import PushStatusTracker, Status


_TOKEN_REFRESH_GRACE_PERIOD = 10 * 60  # 10 minutes


_LOGGER = logging.getLogger(__name__)


class PushManager(object):  # pylint:disable=too-many-instance-attributes
    """Push notifications susbsytem manager."""

    def __init__(self, auth_api, synchronizer, feedback_loop, sse_url=None):
        """
        Class constructor.

        :param auth_api: sdk-auth-service api client
        :type auth_api: splitio.api.auth.AuthAPI

        :param synchronizer: split data synchronizer facade
        :type synchronizer: splitio.sync.synchronizer.Synchronizer

        :param feedback_loop: queue where push status updates are published.
        :type feedback_loop: queue.Queue

        :param sse_url: streaming base url.
        :type sse_url: str
        """
        self._auth_api = auth_api
        self._feedback_loop = feedback_loop
        self._processor = MessageProcessor(synchronizer)
        self._status_tracker = PushStatusTracker()
        self._event_handlers = {
            EventType.MESSAGE: self._handle_message,
            EventType.ERROR: self._handle_error
        }

        self._message_handlers = {
            MessageType.UPDATE: self._handle_update,
            MessageType.CONTROL: self._handle_control,
            MessageType.OCCUPANCY: self._handle_occupancy
        }

        kwargs = {} if sse_url is None else {'base_url': sse_url}
        self._sse_client = SplitSSEClient(self._event_handler, self._handle_connection_ready,
                                          self._handle_connection_end, **kwargs)
        self._running = False
        self._next_refresh = Timer(0, lambda: 0)

    def update_workers_status(self, enabled):
        """
        Enable/Disable push update workers.

        :param enabled: if True, enable workers. If False, disable them.
        :type enabled: bool
        """
        self._processor.update_workers_status(enabled)


    def start(self):
        """Start a new connection if not already running."""
        if self._running:
            _LOGGER.warning('Push manager already has a connection running. Ignoring')
            return

        self._trigger_connection_flow()

    def stop(self, blocking=False):
        """
        Stop the current ongoing connection.

        :param blocking: whether to wait for the connection to be successfully closed or not
        :type blocking: bool
        """
        if not self._running:
            _LOGGER.warning('Push manager does not have an open SSE connection. Ignoring')
            return

        self._running = False
        self._processor.update_workers_status(False)
        self._status_tracker.notify_sse_shutdown_expected()
        self._next_refresh.cancel()
        self._sse_client.stop(blocking)

    def _event_handler(self, event):
        """
        Process an incoming event.

        :param event: Incoming event
        :type event: splitio.push.sse.SSEEvent
        """
        try:
            parsed = parse_incoming_event(event)
        except EventParsingException:
            _LOGGER.error('error parsing event of type %s', event.event_type)
            _LOGGER.debug(str(event), exc_info=True)
            return

        try:
            handle = self._event_handlers[parsed.event_type]
        except KeyError:
            _LOGGER.error('no handler for message of type %s', parsed.event_type)
            _LOGGER.debug(str(event), exc_info=True)
            return

        try:
            handle(parsed)
        except Exception:  #pylint:disable=broad-except
            _LOGGER.error('something went wrong when processing message of type %s',
                          parsed.event_type)
            _LOGGER.debug(str(parsed), exc_info=True)

    def _token_refresh(self):
        """Refresh auth token."""
        _LOGGER.info("retriggering authentication flow.")
        self.stop(True)
        self._trigger_connection_flow()

    def _trigger_connection_flow(self):
        """Authenticate and start a connection."""
        try:
            token = self._auth_api.authenticate()
        except APIException:
            _LOGGER.error('error performing sse auth request.')
            _LOGGER.debug('stack trace: ', exc_info=True)
            self._feedback_loop.put(Status.PUSH_RETRYABLE_ERROR)
            return

        if not token.push_enabled:
            self._feedback_loop.put(Status.PUSH_NONRETRYABLE_ERROR)
            return

        _LOGGER.debug("auth token fetched. connecting to streaming.")
        self._status_tracker.reset()
        if self._sse_client.start(token):
            _LOGGER.debug("connected to streaming, scheduling next refresh")
            self._setup_next_token_refresh(token)
            self._running = True

    def _setup_next_token_refresh(self, token):
        """
        Schedule next token refresh.

        :param token: Last fetched token.
        :type token: splitio.models.token.Token
        """
        if self._next_refresh is not None:
            self._next_refresh.cancel()
        self._next_refresh = Timer((token.exp - token.iat) - _TOKEN_REFRESH_GRACE_PERIOD,
                                   self._token_refresh)
        self._next_refresh.setName('TokenRefresh')
        self._next_refresh.start()

    def _handle_message(self, event):
        """
        Handle incoming update message.

        :param event: Incoming Update message
        :type event: splitio.push.sse.parser.Update
        """
        try:
            handle = self._message_handlers[event.message_type]
        except KeyError:
            _LOGGER.error('no handler for message of type %s', event.message_type)
            _LOGGER.debug(str(event), exc_info=True)
            return

        handle(event)

    def _handle_update(self, event):
        """
        Handle incoming update message.

        :param event: Incoming Update message
        :type event: splitio.push.sse.parser.Update
        """
        _LOGGER.debug('handling update event: %s', str(event))
        self._processor.handle(event)

    def _handle_control(self, event):
        """
        Handle incoming control message.

        :param event: Incoming control message.
        :type event: splitio.push.sse.parser.ControlMessage
        """
        _LOGGER.debug('handling control event: %s', str(event))
        feedback = self._status_tracker.handle_control_message(event)
        if feedback is not None:
            self._feedback_loop.put(feedback)

    def _handle_occupancy(self, event):
        """
        Handle incoming notification message.

        :param event: Incoming occupancy message.
        :type event: splitio.push.sse.parser.Occupancy
        """
        _LOGGER.debug('handling occupancy event: %s', str(event))
        feedback = self._status_tracker.handle_occupancy(event)
        if feedback is not None:
            self._feedback_loop.put(feedback)

    def _handle_error(self, event):
        """
        Handle incoming error message.

        :param event: Incoming ably error
        :type event: splitio.push.sse.parser.AblyError
        """
        _LOGGER.debug('handling ably error event: %s', str(event))
        feedback = self._status_tracker.handle_ably_error(event)
        if feedback is not None:
            self._feedback_loop.put(feedback)

    def _handle_connection_ready(self):
        """Handle a successful connection to SSE."""
        self._feedback_loop.put(Status.PUSH_SUBSYSTEM_UP)
        _LOGGER.info('sse initial event received. enabling')

    def _handle_connection_end(self):
        """
        Handle a connection ending.

        If the connection shutdown was not requested, trigger a restart.
        """
        feedback = self._status_tracker.handle_disconnect()
        if feedback is not None:
            self._feedback_loop.put(feedback)



"""
Event DTO and Storage classes.

The dto is implemented as a namedtuple for performance matters.
"""

from __future__ import print_function
from collections import namedtuple


Event = namedtuple('Event', [
    'key',
    'traffic_type_name',
    'event_type_id',
    'value',
    'timestamp',
    'properties',
])

EventWrapper = namedtuple('EventWrapper', [
    'event',
    'size',
])



"""Split partition module."""

from future.utils import python_2_unicode_compatible


class Partition(object):
    """Partition object class."""

    def __init__(self, treatment, size):
        """
        Class constructor.

        :param treatment: The treatment for the partition
        :type treatment: str
        :param size: A number between 0 a 100
        :type size: float
        """
        if size < 0 or size > 100:
            raise ValueError('size MUST BE between 0 and 100')

        self._treatment = treatment
        self._size = size

    @property
    def treatment(self):
        """Return the treatment associated with this partition."""
        return self._treatment

    @property
    def size(self):
        """Return the percentage owned by this partition."""
        return self._size

    def to_json(self):
        """Return a JSON representation of a partition."""
        return {
            'treatment': self._treatment,
            'size': self._size
        }

    @python_2_unicode_compatible
    def __str__(self):
        """Return string representation of a partition."""
        return '{size}%:{treatment}'.format(size=self._size,
                                            treatment=self._treatment)


def from_raw(raw_partition):
    """
    Build a partition object from a splitChanges partition portion.

    :param raw_partition: JSON snippet of a partition.
    :type raw_partition: dict

    :return: New partition object.
    :rtype: Partition
    """
    return Partition(raw_partition['treatment'], raw_partition['size'])



"""Impressions API module."""

import logging
from itertools import groupby

from future.utils import raise_from

from splitio.api import APIException, headers_from_metadata
from splitio.api.client import HttpClientException
from splitio.engine.impressions import ImpressionsMode


_LOGGER = logging.getLogger(__name__)


class ImpressionsAPI(object):  # pylint: disable=too-few-public-methods
    """Class that uses an httpClient to communicate with the impressions API."""

    def __init__(self, client, apikey, sdk_metadata, mode=ImpressionsMode.OPTIMIZED):
        """
        Class constructor.

        :param client: HTTP Client responsble for issuing calls to the backend.
        :type client: HttpClient
        :param apikey: User apikey token.
        :type apikey: string
        """
        self._client = client
        self._apikey = apikey
        self._metadata = headers_from_metadata(sdk_metadata)
        self._metadata['SplitSDKImpressionsMode'] = mode.name

    @staticmethod
    def _build_bulk(impressions):
        """
        Build an impression bulk formatted as the API expects it.

        :param impressions: List of impressions to bundle.
        :type impressions: list(splitio.models.impressions.Impression)

        :return: Dictionary of lists of impressions.
        :rtype: list
        """
        return [
            {
                'f': test_name,
                'i': [
                    {
                        'k': impression.matching_key,
                        't': impression.treatment,
                        'm': impression.time,
                        'c': impression.change_number,
                        'r': impression.label,
                        'b': impression.bucketing_key,
                        'pt': impression.previous_time
                    }
                    for impression in imps
                ]
            }
            for (test_name, imps) in groupby(
                sorted(impressions, key=lambda i: i.feature_name),
                lambda i: i.feature_name
            )
        ]

    @staticmethod
    def _build_counters(counters):
        """
        Build an impression bulk formatted as the API expects it.

        :param counters: List of impression counters per feature.
        :type counters: list[splitio.engine.impressions.Counter.CountPerFeature]

        :return: dict with list of impression count dtos
        :rtype: dict
        """
        return {
            'pf': [
                {
                    'f': pf_count.feature,
                    'm': pf_count.timeframe,
                    'rc': pf_count.count
                } for pf_count in counters
            ]
        }

    def flush_impressions(self, impressions):
        """
        Send impressions to the backend.

        :param impressions: Impressions bulk
        :type impressions: list
        """
        bulk = self._build_bulk(impressions)
        try:
            response = self._client.post(
                'events',
                '/testImpressions/bulk',
                self._apikey,
                body=bulk,
                extra_headers=self._metadata
            )
            if not 200 <= response.status_code < 300:
                raise APIException(response.body, response.status_code)
        except HttpClientException as exc:
            _LOGGER.error(
                'Error posting impressions because an exception was raised by the HTTPClient'
            )
            _LOGGER.debug('Error: ', exc_info=True)
            raise_from(APIException('Impressions not flushed properly.'), exc)

    def flush_counters(self, counters):
        """
        Send impressions to the backend.

        :param impressions: Impressions bulk
        :type impressions: list
        """
        bulk = self._build_counters(counters)
        try:
            response = self._client.post(
                'events',
                '/testImpressions/count',
                self._apikey,
                body=bulk,
                extra_headers=self._metadata
            )
            if not 200 <= response.status_code < 300:
                raise APIException(response.body, response.status_code)
        except HttpClientException as exc:
            _LOGGER.error(
                'Error posting impressions counters because an exception was raised by the '
                'HTTPClient'
            )
            _LOGGER.debug('Error: ', exc_info=True)
            raise_from(APIException('Impressions not flushed properly.'), exc)
